# -*- coding: utf-8 -*-
#
# This file is part of the qpageview package.
#
# Copyright (c) 2016 - 2019 by Wilbert Berendsen
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 2
# of the License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
# See http://www.gnu.org/licenses/ for more information.

"""
ScrollArea, that supports kinetic scrolling and other features.
"""

import math

from PyQt5.QtCore import QPoint, QRect, QSize, Qt
from PyQt5.QtWidgets import QAbstractScrollArea

from . import util


class ScrollArea(QAbstractScrollArea):
    """A scroll area that supports kinetic scrolling and other features."""

    #: how to align the scrolled area if smaller than the viewport (Qt.AlignCenter)
    alignment = Qt.AlignCenter

    #: how many scroll updates to draw per second (50, 50 is recommended).
    scrollupdatespersec = 50

    #: whether the mouse wheel and PgUp/PgDn keys etc use kinetic scrolling (True)
    kineticScrollingEnabled = True

    #: If enabled, the user can drag the contents of the scrollarea to
    #: move it with the mouse.
    draggingEnabled = True

    def __init__(self, parent=None, **kwds):
        super().__init__(parent, **kwds)
        self._areaSize = 0, 0
        self._dragPos = None
        self._dragSpeed = None
        self._dragTime = None
        self._scroller = None
        self._scrollTimer = None

    def setAreaSize(self, size):
        """Updates the scrollbars to be able to display an area of this size."""
        self._areaSize = (size.width(), size.height())
        self._updateScrollBars()

    def areaSize(self):
        """Return the size of the area as set by setAreaSize()."""
        return QSize(*self._areaSize)

    def areaPos(self):
        """Return the position of the area relative to the viewport.

        The alignment attribute is taken into account when the area is smaller
        than the viewport (horizontally and/or vertically).

        """
        w, h = self._areaSize
        vw = self.viewport().width()
        vh = self.viewport().height()
        left, top = util.align(w, h, vw, vh, self.alignment)
        if left < 0:
            left = -self.horizontalScrollBar().value()
        if top < 0:
            top = -self.verticalScrollBar().value()
        return QPoint(left, top)

    def visibleArea(self):
        """Return a rectangle describing the part of the area that is visible."""
        pos = self.areaPos()
        r = self.viewport().rect() & QRect(pos, self.areaSize())
        return r.translated(-pos)

    def offsetToEnsureVisible(self, rect):
        """Return an offset QPoint with the minimal scroll to make rect visible.

        If the rect is too large, it is positioned top-left.

        """
        area = self.visibleArea()
        # vertical
        dy = 0
        if rect.bottom() > area.bottom():
            dy = rect.bottom() - area.bottom()
        if rect.top() < area.top() + dy:
            dy = rect.top() - area.top()
        # horizontal
        dx = 0
        if rect.right() > area.right():
            dx = rect.right() - area.right()
        if rect.left() < area.left() + dx:
            dx = rect.left() - area.left()
        return QPoint(dx, dy)

    def ensureVisible(self, rect, margins=None, allowKinetic=True):
        """Performs the minimal scroll to make rect visible.

        If the rect is not completely visible it is scrolled into view, adding
        the margins if given (a QMargins instance). If allowKinetic is False,
        immediately jumps to the position, otherwise scrolls smoothly (if
        kinetic scrolling is enabled).

        """
        if rect not in self.visibleArea():
            if margins is not None:
                rect = rect + margins
            diff = self.offsetToEnsureVisible(rect)
            if allowKinetic and self.kineticScrollingEnabled:
                self.kineticScrollBy(diff)
            else:
                self.scrollBy(diff)

    def _updateScrollBars(self):
        """Internal. Adjust the range of the scrollbars to the area size.

        Called in setAreaSize() and resizeEvent().

        """
        w, h = self._areaSize
        maxsize = self.maximumViewportSize()
        vbar = self.verticalScrollBar()
        hbar = self.horizontalScrollBar()

        if w <= maxsize.width() and h <= maxsize.height():
            vbar.setRange(0, 0)
            hbar.setRange(0, 0)
        else:
            viewport = self.viewport()
            vbar.setRange(0, h - viewport.height())
            vbar.setPageStep(int(viewport.height() * .9))
            hbar.setRange(0, w - viewport.width())
            hbar.setPageStep(int(viewport.width() * .9))

    def scrollOffset(self):
        """Return the current scroll offset."""
        x = self.horizontalScrollBar().value()
        y = self.verticalScrollBar().value()
        return QPoint(x, y)

    def canScrollBy(self, diff):
        """Does not scroll, but return the actual distance the View would scroll.

        diff is a QPoint instance.

        """
        hbar = self.horizontalScrollBar()
        vbar = self.verticalScrollBar()

        x = min(max(0, hbar.value() + diff.x()), hbar.maximum())
        y = min(max(0, vbar.value() + diff.y()), vbar.maximum())
        return QPoint(x - hbar.value(), y - vbar.value())

    def scrollForDragging(self, pos):
        """Slowly scroll the View if pos is close to the edge of the viewport.

        Can be used while dragging things.

        """
        viewport = self.viewport().rect()
        dx = pos.x() - viewport.left() - 12
        if dx >= 0:
            dx = max(0, pos.x() - viewport.right() + 12)
        dy = pos.y() - viewport.top() - 12
        if dy >= 0:
            dy = max(0, pos.y() - viewport.bottom() + 12)
        self.steadyScroll(QPoint(dx*10, dy*10))

    def scrollTo(self, pos):
        """Scroll the View to get pos (QPoint) in the top left corner (if possible).

        Returns the actual distance moved.

        """
        return self.scrollBy(pos - self.scrollOffset())

    def scrollBy(self, diff):
        """Scroll the View diff pixels (QPoint) in x and y direction.

        Returns the actual distance moved.

        """
        hbar = self.horizontalScrollBar()
        vbar = self.verticalScrollBar()
        x = hbar.value()
        hbar.setValue(hbar.value() + diff.x())
        x = hbar.value() - x
        y = vbar.value()
        vbar.setValue(vbar.value() + diff.y())
        y = vbar.value() - y
        return QPoint(x, y)

    def kineticScrollTo(self, pos):
        """Scroll the View to get pos (QPoint) in the top left corner (if possible).

        Returns the actual distance the scroll area will move.

        """
        return self.kineticScrollBy(pos - self.scrollOffset())

    def kineticScrollBy(self, diff):
        """Scroll the View diff pixels (QPoint) in x and y direction.

        Returns the actual distance the scroll area will move.

        """
        ret = self.canScrollBy(diff)
        if diff:
            scroller = KineticScroller()
            scroller.scrollBy(diff)
            self.startScrolling(scroller)
        return ret

    def kineticAddDelta(self, diff):
        """Add diff (QPoint) to an existing kinetic scroll.

        If no scroll is active, a new one is started (like kineticScrollBy).

        """
        if isinstance(self._scroller, KineticScroller):
            self._scroller.scrollBy(self._scroller.remainingDistance() + diff)
        else:
            self.kineticScrollBy(diff)

    def steadyScroll(self, diff):
        """Start steadily scrolling diff (QPoint) pixels per second.

        Stops automatically when the end is reached.

        """
        if diff:
            self.startScrolling(SteadyScroller(diff, self.scrollupdatespersec))
        else:
            self.stopScrolling()

    def startScrolling(self, scroller):
        """Begin a scrolling operation using the specified scroller."""
        self._scroller = scroller
        if self._scrollTimer is None:
            self._scrollTimer = self.startTimer(1000 // self.scrollupdatespersec)

    def stopScrolling(self):
        """Stop scrolling."""
        if self._scroller:
            self.killTimer(self._scrollTimer)
            self._scroller = None
            self._scrollTimer = None

    def isScrolling(self):
        """Return True if a scrolling movement is active."""
        return self._scroller is not None

    def remainingScrollTime(self):
        """If a kinetic scroll is active, return how many msecs the scroll wil last.

        Otherwise, return 0.

        """
        if isinstance(self._scroller, KineticScroller):
            return 1000 // self.scrollupdatespersec * self._scroller.remainingTicks()
        return 0

    def isDragging(self):
        """Return True if the user is dragging the background."""
        return self._dragPos is not None

    def timerEvent(self, ev):
        """Implemented to handle the scroll timer."""
        if ev.timerId() == self._scrollTimer:
            diff = self._scroller.step()
            # when scrolling slowly, it might be that no redraw is needed
            if diff:
                # change the scrollbars, but check how far they really moved.
                if not self.scrollBy(diff) or self._scroller.finished():
                    self.stopScrolling()

    def resizeEvent(self, ev):
        """Implemented to update the scrollbars to the aera size."""
        self._updateScrollBars()

    def mousePressEvent(self, ev):
        """Implemented to handle dragging the document with the left button."""
        self.stopScrolling()
        super().mousePressEvent(ev)

    def mouseMoveEvent(self, ev):
        """Implemented to handle dragging the document with the left button."""
        if self.draggingEnabled and ev.buttons() & Qt.LeftButton:
            if self._dragPos is None:
                self.setCursor(Qt.SizeAllCursor)
            else:
                diff = self._dragPos - ev.pos()
                self._dragSpeed = (ev.timestamp() - self._dragTime, diff)
                self.scrollBy(diff)
            self._dragPos = ev.pos()
            self._dragTime = ev.timestamp()
        super().mouseMoveEvent(ev)

    def mouseReleaseEvent(self, ev):
        """Implemented to handle dragging the document with the left button."""
        if self.draggingEnabled and ev.button() == Qt.LeftButton and self._dragPos is not None:
            self.unsetCursor()
            if self.kineticScrollingEnabled and self._dragSpeed is not None:
                # compute speed of last movement
                time, speed = self._dragSpeed
                time += ev.timestamp() - self._dragTime # add time between last mvt and release
                speed = speed * 1000 / self.scrollupdatespersec / time
                # compute diff to scroll
                sx = abs(speed.x())
                diffx = int(sx * (sx + 1) / 2)
                sy = abs(speed.y())
                diffy = int(sy * (sy + 1) / 2)
                if speed.x() < 0: diffx = -diffx
                if speed.y() < 0: diffy = -diffy
                self.kineticScrollBy(QPoint(diffx, diffy))
            self._dragPos = None
            self._dragTime = None
            self._dragSpeed = None
        super().mouseReleaseEvent(ev)

    def wheelEvent(self, ev):
        """Reimplemented to use kinetic mouse wheel scrolling if enabled."""
        if self.kineticScrollingEnabled:
            self.kineticAddDelta(-ev.angleDelta())
        else:
            super().wheelEvent(ev)

    def keyPressEvent(self, ev):
        """Reimplemented to use kinetic cursor movements."""
        hbar = self.horizontalScrollBar()
        vbar = self.verticalScrollBar()
        # add Home and End, even in non-kinetic mode
        scroll = self.kineticScrollBy if self.kineticScrollingEnabled else self.scrollBy
        if ev.key() == Qt.Key_Home:
            scroll(QPoint(0, -vbar.value()))
        elif ev.key() == Qt.Key_End:
            scroll(QPoint(0, vbar.maximum() - vbar.value()))
        elif self.kineticScrollingEnabled:
            # make arrow keys and PgUp and PgDn kinetic
            if ev.key() == Qt.Key_PageDown:
                self.kineticAddDelta(QPoint(0, vbar.pageStep()))
            elif ev.key() == Qt.Key_PageUp:
                self.kineticAddDelta(QPoint(0, -vbar.pageStep()))
            elif ev.key() == Qt.Key_Down:
                self.kineticAddDelta(QPoint(0, vbar.singleStep()))
            elif ev.key() == Qt.Key_Up:
                self.kineticAddDelta(QPoint(0, -vbar.singleStep()))
            elif ev.key() == Qt.Key_Left:
                self.kineticAddDelta(QPoint(-hbar.singleStep(), 0))
            elif ev.key() == Qt.Key_Right:
                self.kineticAddDelta(QPoint(hbar.singleStep(), 0))
            else:
                super().keyPressEvent(ev)
        else:
            super().keyPressEvent(ev)


class Scroller:
    """Abstract base class, encapsulates scrolling behaviour.

    A Scroller subclass must implement the step() and finished() methods and
    may define additional methods.

    """
    def step(self):
        """Implement this method to return a QPoint for the current scrolling step."""

    def finished(self):
        """Implement this method to return True if scrolling is finished."""


class SteadyScroller(Scroller):
    """Scrolls the area steadily n pixels per second."""
    def __init__(self, speed, updates_per_second):
        """Initializes with speed (QPoint) pixels per second."""
        self._x = speed.x()
        self._y = speed.y()
        self._restx = 0
        self._resty = 0
        self._ups = updates_per_second

    def step(self):
        """Return a QPoint indicating the diff to scroll in this step.

        If this is a QPoint(0, 0) it does not indicate that scrolling has
        finished. Use finished() for that.

        """
        # the amount of pixels to scroll per second
        x = self._x
        y = self._y

        # how many updates per second, compute the number of pixes to scroll now
        ups = self._ups
        dx, rx = divmod(abs(x), ups)
        dy, ry = divmod(abs(y), ups)
        dx1, self._restx = divmod(self._restx + rx, ups)
        dy1, self._resty = divmod(self._resty + ry, ups)
        dx += dx1
        dy += dy1

        # scroll in the right direction
        diff = QPoint(-dx if x < 0 else dx, -dy if y < 0 else dy)
        return diff

    def finished(self):
        """As this scroller has a constant speed, it never stops."""
        return False


class KineticScroller(Scroller):
    """Scrolls the area with a decreasing speed."""
    def __init__(self):
        self._x = 0
        self._y = 0
        self._offset = None

    def scrollBy(self, diff):
        """Start a new kinetic scroll of the specified amount."""
        ### logic by Richard Cognot, May 2012, simplified by WB
        dx = diff.x()
        dy = diff.y()

        # solve speed*(speed+1)/2 = delta to ensure 1+2+3+...+speed is as close as possible under delta..
        sx = int(math.sqrt(1 + 8 * abs(dx)) - 1) // 2
        sy = int(math.sqrt(1 + 8 * abs(dy)) - 1) // 2

        # compute the amount of displacement still needed because we're dealing with integer values.
        # Since this function is called for exact moves (not free scrolling)
        # limit the kinetic time to 2 seconds, which means 100 ticks, 5050 pixels.
        # (TODO: adapt for other ticker speeds? WB)
        if sy > 100:
            sy = 100
        offy = abs(dy) - sy * (sy + 1) // 2

        # Although it is less likely to go beyond that limit for horizontal scrolling,
        # do it for x as well.
        if sx > 100:
            sx = 100
        offx = abs(dx) - sx * (sx + 1) // 2

        # adjust directions
        if dx < 0:
            sx = -sx
            offx = -offx
        if dy < 0:
            sy = -sy
            offy = -offy
        self._x = sx
        self._y = sy
        # the offset is accounted for in the first step
        self._offset = QPoint(offx, offy)

    def remainingDistance(self):
        """Return the remaining distance."""
        sx = abs(self._x)
        dx = sx * (sx + 1) // 2
        if self._x < 0:
            dx = -dx
        sy = abs(self._y)
        dy = sy * (sy + 1) // 2
        if self._y < 0:
            dy = -dy
        return QPoint(dx, dy)

    def remainingTicks(self):
        """Return the remaining ticks of this scroll."""
        return max(abs(self._x), abs(self._y))

    def step(self):
        """Return a QPoint indicating the diff to scroll in this step."""
        ret = QPoint(self._x, self._y)
        if self._offset:
            ret += self._offset
            self._offset = None
        if self._x > 0:
            self._x -= 1
        elif self._x < 0:
            self._x += 1
        if self._y > 0:
            self._y -= 1
        elif self._y < 0:
            self._y += 1
        return ret

    def finished(self):
        """Return True if scrolling is done."""
        return self._x == 0 and self._y == 0





import sys

sys.path.insert(0, "../source")
from conf import *


extensions.extend([
    "sphinx_rtd_theme",
])

html_theme = "sphinx_rtd_theme"
html_theme_path = [
    # we need the current git version of the theme, pip3 install fails...
    "/home/wilbert/src/sphinx_rtd_theme/",
]

html_logo = '_static/qp.png'

html_theme_options = {
    'canonical_url': 'https://qpageview.org/',
#    'analytics_id': 'UA-XXXXXXX-1',  #  Provided by Google in your dashboard
    'logo_only': True,
    'display_version': True,
    'prev_next_buttons_location': 'bottom',
    'style_external_links': False,
    #'style_nav_header_background': 'white',
    'style_nav_header_background': 'firebrick',
    # Toc options
    'collapse_navigation': True,
    'sticky_navigation': True,
    'navigation_depth': 4,
    'includehidden': True,
    'titles_only': False,
}

html_context = {
  'display_github': True,
  'github_user': 'frescobaldi',
  'github_repo': 'qpageview',
  'github_version': 'master',
  'conf_py_path': '/docs/source/',
}




# -*- coding: utf-8 -*-
#
# This file is part of the qpageview package.
#
# Copyright (c) 2016 - 2019 by Wilbert Berendsen
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 2
# of the License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
# See http://www.gnu.org/licenses/ for more information.

"""
A page that can display a SVG document.

"""

from PyQt5.QtCore import QRect, QRectF, Qt
from PyQt5.QtGui import QColor, QPainter
from PyQt5.QtSvg import QSvgRenderer

from .constants import (
    Rotate_0,
    Rotate_90,
    Rotate_180,
    Rotate_270,
)

from . import document
from . import locking
from . import page
from . import render


class SvgPage(page.AbstractRenderedPage):
    """A page that can display a SVG document."""

    dpi = 90.0

    def __init__(self, svgrenderer, renderer=None):
        super().__init__(renderer)
        self._svg = svgrenderer
        self.pageWidth = svgrenderer.defaultSize().width()
        self.pageHeight = svgrenderer.defaultSize().height()
        self._viewBox = svgrenderer.viewBoxF()

    @classmethod
    def load(cls, filename, renderer=None):
        """Load a SVG document from filename, which may also be a QByteArray.

        Yields only one Page instance, as SVG currently supports one page per
        file. If the file can't be loaded by the underlying QSvgRenderer,
        no Page is yielded.

        """
        r = QSvgRenderer()
        if r.load(filename):
            yield cls(r, renderer)

    def mutex(self):
        return self._svg

    def group(self):
        return self._svg


class SvgDocument(document.MultiSourceDocument):
    """A Document representing a group of SVG files."""
    pageClass = SvgPage

    def createPages(self):
        return self.pageClass.loadFiles(self.sources(), self.renderer)


class SvgRenderer(render.AbstractRenderer):
    """Render SVG pages."""
    def setRenderHints(self, painter):
        """Sets the renderhints for the painter we want to use."""
        painter.setRenderHint(QPainter.Antialiasing, self.antialiasing)
        painter.setRenderHint(QPainter.TextAntialiasing, self.antialiasing)

    def draw(self, page, painter, key, tile, paperColor=None):
        """Draw the specified tile of the page (coordinates in key) on painter."""
        # determine the part to draw; convert tile to viewbox
        viewbox = self.map(key, page._viewBox).mapRect(QRectF(*tile))
        target = QRectF(0, 0, tile.w, tile.h)
        if key.rotation & 1:
            target.setSize(target.size().transposed())
        with locking.lock(page._svg):
            page._svg.setViewBox(viewbox)
            # we must specify the target otherwise QSvgRenderer scales to the
            # unrotated image
            painter.save()
            painter.setClipRect(target, Qt.IntersectClip)
            # QSvgRenderer seems to set antialiasing always on anyway... :-)
            self.setRenderHints(painter)
            page._svg.render(painter, target)
            painter.restore()
            page._svg.setViewBox(page._viewBox)




# install a default renderer, so SvgPage can be used directly
SvgPage.renderer = SvgRenderer()







# -*- coding: utf-8 -*-
#
# This file is part of the qpageview package.
#
# Copyright (c) 2010 - 2019 by Wilbert Berendsen
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 2
# of the License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
# See http://www.gnu.org/licenses/ for more information.


"""
Rubberband selection in a View.
"""


from PyQt5.QtCore import QEvent, QPoint, QRect, QSize, Qt, pyqtSignal
from PyQt5.QtGui import QColor, QContextMenuEvent, QCursor, QPainter, QPalette, QPen, QRegion
from PyQt5.QtWidgets import QApplication, QWidget


# dragging/moving selection:
_OUTSIDE = 0
_LEFT    = 1
_TOP     = 2
_RIGHT   = 4
_BOTTOM  = 8
_INSIDE  = 15


class Rubberband(QWidget):
    """A Rubberband to select a rectangular region.

    A Rubberband is added to a View with view.setRubberband().

    The Rubberband lets the user select a rectangular region. When the
    selection is changed, the `selectionChanged` signal is emitted, having the
    selection rectangle in layout coordinates as argument.

    Instance variables:

    ``showbutton`` (Qt.RightButton)
        the button used to drag a new rectangle

    ``dragbutton`` (Qt.LeftButton)
        the button to alter an existing rectangle

    ``trackSelection`` (False)
        whether to continuously emit selectionChanged(). When True,
        ``selectionChanged()`` is emitted on every change, when False, the signal
        is only emitted when the mouse button is released.

    """
    selectionChanged = pyqtSignal(QRect)

    # the button used to drag a new rectangle
    showbutton = Qt.RightButton

    # the button to alter an existing rectangle
    dragbutton = Qt.LeftButton

    # whether to continuously track the selection
    trackSelection = False

    def __init__(self):
        super().__init__()
        self._dragging = False
        self._dragedge = 0
        self._dragpos = None
        self._selection = QRect()
        self._layoutOffset = None   # used to keep on spot during resize/zoom
        self.setMouseTracking(True)
        self.setContextMenuPolicy(Qt.PreventContextMenu)

    def paintEvent(self, ev):
        ### Paint code contributed by Richard Cognot Jun 2012
        color = self.palette().color(QPalette.Highlight)
        painter = QPainter(self)

        # Filled rectangle.
        painter.setClipRect(self.rect())
        color.setAlpha(50)
        painter.fillRect(self.rect().adjusted(2,2,-2,-2), color)

        # Thin rectangle outside.
        color.setAlpha(150)
        painter.setPen(color)
        # XXX can this adjustment be done smarter?
        adjust = int(-1 / self.devicePixelRatio())
        painter.drawRect(self.rect().adjusted(0, 0, adjust, adjust))

        # Pseudo-handles at the corners and sides
        color.setAlpha(100)
        pen = QPen(color)
        pen.setWidth(8)
        painter.setPen(pen)
        painter.setBackgroundMode(Qt.OpaqueMode)
        # Clip at 4 corners
        region = QRegion(QRect(0,0,20,20))
        region += QRect(self.rect().width()-20, 0, 20, 20)
        region += QRect(self.rect().width()-20, self.rect().height()-20, 20, 20)
        region += QRect(0, self.rect().height()-20, 20, 20)
        # Clip middles
        region += QRect(0, self.rect().height() // 2 - 10, self.rect().width(), 20)
        region += QRect(self.rect().width() // 2 - 10, 0, 20, self.rect().height())

        # Draw thicker rectangles, clipped at corners and sides.
        painter.setClipRegion(region)
        painter.drawRect(self.rect())

    def edge(self, point):
        """Return the edge where the point touches our geometry."""
        rect = self.geometry()
        if point not in rect:
            return _OUTSIDE
        edge = 0
        if point.x() <= rect.left() + 8:
            edge |= _LEFT
        elif point.x() >= rect.right() - 8:
            edge |= _RIGHT
        if point.y() <= rect.top() + 8:
            edge |= _TOP
        elif point.y() >= rect.bottom() - 8:
            edge |= _BOTTOM
        return edge or _INSIDE

    def adjustCursor(self, edge):
        """Sets the cursor shape when we are at edge."""
        cursor = None
        if edge in (_TOP, _BOTTOM):
            cursor = Qt.SizeVerCursor
        elif edge in (_LEFT, _RIGHT):
            cursor = Qt.SizeHorCursor
        elif edge in (_LEFT | _TOP, _RIGHT | _BOTTOM):
            cursor = Qt.SizeFDiagCursor
        elif edge in (_TOP | _RIGHT, _BOTTOM | _LEFT):
            cursor = Qt.SizeBDiagCursor
        elif edge is _INSIDE:
            cursor = Qt.SizeAllCursor
        if cursor:
            self.setCursor(cursor)
        else:
            self.unsetCursor()

    def hasSelection(self):
        """Return True when there is a selection."""
        return bool(self._selection)

    def selection(self):
        """Return our selection rectangle, relative to the view's layout position."""
        return self._selection

    def selectedPages(self):
        """Yield tuples (page, rect) describing the selection.

        Every rect is intersected with the page rect and translated to the
        page's position.

        """
        rect = self.selection()
        if rect:
            view = self.parent().parent()
            layout = view.pageLayout()
            for page in layout.pagesAt(rect):
                yield page, rect.intersected(page.geometry()).translated(-page.pos())

    def selectedPage(self):
        """Returns (page, rect) if there is a selection.

        If the selection contains more pages, the largest intersection is chosen.
        If no meaningful area is selected, (None, None) is returned.

        """
        selection = sorted(self.selectedPages(), key=lambda pr: pr[1].height() + pr[1].width())
        if selection:
            return selection[-1]
        else:
            return None, None

    def selectedImage(self, resolution=None, paperColor=None):
        """Returns an image of the selected part on a Page.

        If resolution is None, the displayed size is chosen. Otherwise, the
        resolution is an integer, interpreted as DPI (dots per inch).

        """
        page, rect = self.selectedPage()
        if page and rect:
            if resolution is None:
                view = self.parent().parent()
                try:
                    ratio = view.devicePixelRatioF()
                except AttributeError:
                    ratio = view.devicePixelRatio()
                resolution = view.physicalDpiX() * view.zoomFactor() * ratio
            return page.image(rect, resolution, resolution, paperColor)

    def selectedText(self):
        """Return the text found in the selection, as far as the pages support it."""
        result = []
        for page, rect in self.selectedPages():
            result.append(page.text(rect))
        return '\n'.join(result)

    def selectedLinks(self):
        """Yield tuples (page, links) for every page in the selection.

        links is a non-empty set() of Link instances on that page that intersect
        with the selection.

        """
        for page, rect in self.selectedPages():
            links = page.linksIn(rect)
            if links:
                yield page, links

    def setSelection(self, rect):
        """Sets the selection, the rectangle should be relative to the view's layout position."""
        if rect:
            view = self.parent().parent()
            geom = rect.translated(view.layoutPosition())
            self.setGeometry(geom)
            self._setLayoutOffset(geom.topLeft())
            self._oldZoom = view.zoomFactor()
            self.show()
            self._setSelectionFromGeometry(geom)
        else:
            self.hide()
            self._setSelectionFromGeometry(QRect())

    def clearSelection(self):
        """Hide ourselves and clear the selection."""
        self.hide()
        self._dragging = False
        self._setSelectionFromGeometry(QRect())

    def _setSelectionFromGeometry(self, rect):
        """(Internal) Called to emit the selectionChanged signal.

        Only emits the signal when the selection really changed.
        The rect should be our geometry or an empty QRect().

        """
        if rect:
            view = self.parent().parent()
            rect = rect.translated(-view.layoutPosition())
        old, self._selection = self._selection, rect
        if rect != old:
            self.selectionChanged.emit(rect)

    def _setLayoutOffset(self, pos):
        """Store the position as offset from the layout, and also from the page
        at that position. Used for keeping the same spot on zoom change.

        """
        view = self.parent().parent()
        pos = pos - view.layoutPosition()
        self._layoutOffset = view.pageLayout().pos2offset(pos)

    def _getLayoutOffset(self):
        """Get the stored layout offset position back, after zoom or move."""
        view = self.parent().parent()
        pos = view.pageLayout().offset2pos(self._layoutOffset)
        return pos + view.layoutPosition()

    def scrollBy(self, diff):
        """Called by the View when scrolling."""
        if not self._dragging:
            self.move(self.pos() + diff)
            # adjust the cursor
            self.adjustCursor(self.edge(self.parent().mapFromGlobal(QCursor.pos())))
        elif self._dragedge != _INSIDE:
            self._draggeom.moveTo(self._draggeom.topLeft() + diff)
            self.dragBy(-diff)
        elif self.isVisible() and self.trackSelection:
            self._setSelectionFromGeometry(self.geometry())

    def startDrag(self, pos, button):
        """Start dragging the rubberband."""
        self._dragging = True
        self._dragpos = pos
        self._dragedge = self.edge(pos)
        self._draggeom = self.geometry()
        self._dragbutton = button

    def drag(self, pos):
        """Continue dragging the rubberband, scrolling the View if necessary."""
        diff = pos - self._dragpos
        self._dragpos = pos
        self.dragBy(diff)
        # check if we are dragging close to the edge of the view, scroll if needed
        view = self.parent().parent()
        view.scrollForDragging(pos)

    def dragBy(self, diff):
        """Drag by diff (QPoint)."""
        edge = self._dragedge
        self._draggeom.adjust(
            diff.x() if edge & _LEFT   else 0,
            diff.y() if edge & _TOP    else 0,
            diff.x() if edge & _RIGHT  else 0,
            diff.y() if edge & _BOTTOM else 0)
        geom = self._draggeom.normalized()
        if geom.isValid():
            self.setGeometry(geom)
            if self.trackSelection:
                self._setSelectionFromGeometry(geom)
        if self.cursor().shape() in (Qt.SizeBDiagCursor, Qt.SizeFDiagCursor):
            # we're dragging a corner, use correct diagonal cursor
            bdiag = (edge in (3, 12)) ^ (self._draggeom.width() * self._draggeom.height() >= 0)
            self.setCursor(Qt.SizeBDiagCursor if bdiag else Qt.SizeFDiagCursor)

    def stopDrag(self):
        """Stop dragging the rubberband."""
        self._dragging = False
        # TODO: use the kinetic scroller if implemented
        view = self.parent().parent()
        view.stopScrolling()

        if self.width() < 8 and self.height() < 8:
            self.unsetCursor()
            self._setSelectionFromGeometry(QRect())
        else:
            self._setSelectionFromGeometry(self.geometry())
            self._setLayoutOffset(self.pos())

    def slotZoomChanged(self, zoom):
        """Called when the zooming in the view changes, resizes ourselves."""
        if self.hasSelection():
            view = self.parent().parent()
            factor =  zoom / self._oldZoom
            self._oldZoom = zoom
            geom = QRect(self._getLayoutOffset(), self.size() * factor)
            self.setGeometry(geom)
            self._setSelectionFromGeometry(geom)

    def eventFilter(self, viewport, ev):
        """Act on events in the viewport:

        * keep on the same place when the viewport resizes
        * start dragging the selection if showbutton clicked (preventing the
          contextmenu if the showbutton is the right button)
        * end a drag on mousebutton release, if that button would have shown
          the context menu, show it on button release.

        """
        if ev.type() == QEvent.Resize and self.isVisible():
            view = self.parent().parent()
            if not view.viewMode():
                # fixed scale, try to keep ourselves in the same position on resize
                self.move(self._getLayoutOffset())
        elif (self.showbutton == Qt.RightButton and isinstance(ev, QContextMenuEvent)
              and ev.reason() == QContextMenuEvent.Mouse):
            # suppress context menu event if that would coincide with start selection
            if not self._dragging or (self.geometry() and self.edge(ev.pos()) == _INSIDE):
                return False
            return True
        elif not self._dragging:
            if ev.type() == QEvent.MouseButtonPress and ev.button() == self.showbutton:
                if self.isVisible():
                    # this cancels a previous selection if we were visible
                    self._setSelectionFromGeometry(QRect())
                self.setGeometry(QRect(ev.pos(), QSize(0, 0)))
                self._setLayoutOffset(ev.pos())
                self._oldZoom = viewport.parent().zoomFactor()
                self.startDrag(ev.pos(), ev.button())
                self._dragedge = _RIGHT | _BOTTOM
                self.adjustCursor(self._dragedge)
                self.show()
                return True
        elif self._dragging:
            if ev.type() == QEvent.MouseMove:
                self.drag(ev.pos())
                return True
            elif ev.type() == QEvent.MouseButtonRelease and ev.button() == self._dragbutton:
                self.stopDrag()
                if ev.button() == Qt.RightButton:
                    QApplication.postEvent(viewport,
                        QContextMenuEvent(QContextMenuEvent.Mouse, ev.pos()))
                return True
        return False

    def mousePressEvent(self, ev):
        """Can start a new drag when we are clicked ourselves."""
        pos = self.mapToParent(ev.pos())
        if not self._dragging:
            if ev.button() == self.dragbutton:
                self.startDrag(pos, ev.button())
            elif ev.button() == self.showbutton:
                if self.showbutton != Qt.RightButton or self.edge(pos) != _INSIDE:
                    self.startDrag(pos, ev.button())

    def mouseMoveEvent(self, ev):
        """Move if we are dragging; show the correct cursor shape on the edges."""
        pos = self.mapToParent(ev.pos())
        if self._dragging:
            self.drag(pos)
        else:
            edge = self.edge(pos)
            self.adjustCursor(edge)

    def mouseReleaseEvent(self, ev):
        """End a self-initiated drag; if the right button was used; send a context menu event."""
        if self._dragging and ev.button() == self._dragbutton:
            self.stopDrag()
        if ev.button() == Qt.RightButton:
            QApplication.postEvent(self.parent(),
                QContextMenuEvent(QContextMenuEvent.Mouse, ev.pos() + self.pos()))





# -*- coding: utf-8 -*-
#
# This file is part of the qpageview package.
#
# Copyright (c) 2019 - 2019 by Wilbert Berendsen
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 2
# of the License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
# See http://www.gnu.org/licenses/ for more information.

"""
Run jobs in the background using QThread.
"""


from PyQt5.QtCore import QThread

# as soon as a Job start()s, it is saved here, to prevent it being
# destroyed while running
_runningJobs = set()
_pendingJobs = []
maxjobs = 12


class Job(QThread):
    """A simple wrapper around QThread.

    Before calling start() you should put the work function in the work
    attribute, and an optional finalize function (which will be called with the
    result) in the finalize attribute.

    Or alternatively, inherit from this class and implement work() and finish()
    yourself. The result of the work function is stored in the result attribute.

    """
    finalize = None
    running = False
    done = False
    result = None

    def __init__(self, parent=None):
        """Init ourselves; the parent can be a QObject which will be our parent."""
        super().__init__(parent)
        self.finished.connect(self._slotFinished)

    def start(self):
        self.result = None
        self.running = True     # this is more robust than isRunning()
        self.done = False
        if len(_runningJobs) < maxjobs:
            _runningJobs.add(self)
            super().start()
        else:
            _pendingJobs.append(self)

    def run(self):
        """Call the work function in the background thread."""
        self.result = self.work()

    def work(self):
        """Implement this to get the work done.

        If you have long tasks you can Qt's isInterruptionRequested()
        functionality.

        Instead of implementing this method, you can put the work function in
        the work instance attribute.

        """
        pass

    def finish(self):
        """This slot is called in the main thread when the work is done.

        The default implementation calls the finalize function with the result.

        """
        if self.finalize:
            self.finalize(self.result)

    def _slotFinished(self):
        _runningJobs.discard(self)
        if _pendingJobs:
            _pendingJobs.pop().start()
        self.running = False
        self.done = True
        self.finish()


class SingleRun:
    """Run a function in a background thread.

    The outcome is silently discarded if another function is called before the
    old one finishes.

    """
    def __init__(self):
        self._job = None

    def cancel(self):
        """Forgets the running job.

        The job is not terminated but the callback is not called.

        """
        j = self._job
        if j:
            j.finalize = None
            self._job = None

    def __call__(self, func, callback=None):
        self.cancel()
        j = self._job = Job()
        j.work = func
        def finalize(result):
            self.cancel()
            if callback:
                callback(result)
        j.finalize = finalize
        j.start()


def run(func, callback=None):
    """Run specified function in a background thread.

    The thread is immediately started. If a callback is specified, it is called
    in the main thread with the result when the function is ready.

    """
    j = Job()
    j.work = func
    j.finalize = callback
    j.start()





# -*- coding: utf-8 -*-
#
# This file is part of the qpageview package.
#
# Copyright (c) 2016 - 2019 by Wilbert Berendsen
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 2
# of the License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
# See http://www.gnu.org/licenses/ for more information.

"""
Interface with popplerqt5, popplerqt5-specific classes etc.

This module depends on popplerqt5, although it can be imported when
popplerqt5 is not available.

You need this module to display PDF documents.

"""

import contextlib
import weakref

from PyQt5.QtCore import Qt, QRectF
from PyQt5.QtGui import QRegion, QPainter, QPicture, QTransform

try:
    import popplerqt5
except ImportError:
    popplerqt5 = None

from . import document
from . import page
from . import link
from . import locking
from . import render

from .constants import (
    Rotate_0,
    Rotate_90,
    Rotate_180,
    Rotate_270,
)



# store the links in the page of a Poppler document as long as the document exists
_linkscache = weakref.WeakKeyDictionary()


class Link(link.Link):
    """A Link that encapsulates a Poppler.Link object."""
    def __init__(self, linkobj):
        self.linkobj = linkobj
        self.area = link.Area(*linkobj.linkArea().normalized().getCoords())

    @property
    def url(self):
        """The url the link points to."""
        if isinstance(self.linkobj, popplerqt5.Poppler.LinkBrowse):
            return self.linkobj.url()
        return ""


class PopplerPage(page.AbstractRenderedPage):
    """A Page capable of displaying one page of a Poppler.Document instance.

    It has two additional instance attributes:

        `document`: the Poppler.Document instance
        `pageNumber`: the page number to render

    """
    def __init__(self, document, pageNumber, renderer=None):
        super().__init__(renderer)
        self.document = document
        self.pageNumber = pageNumber
        self.setPageSize(document.page(pageNumber).pageSizeF())

    @classmethod
    def loadPopplerDocument(cls, document, renderer=None, pageSlice=None):
        """Convenience class method yielding instances of this class.

        The Page instances are created from the document, in page number order.
        The specified Renderer is used, or else the global poppler renderer.
        If pageSlice is given, it should be a slice object and only those pages
        are then loaded.

        """
        it = range(document.numPages())
        if pageSlice is not None:
            it = it[pageSlice]
        for num in it:
            yield cls(document, num, renderer)

    @classmethod
    def load(cls, filename, renderer=None):
        """Load a Poppler document, and yield of instances of this class.

        The filename can also be a QByteArray or a popplerqt5.Poppler.Document
        instance. The specified Renderer is used, or else the global poppler
        renderer.

        """
        doc = load(filename)
        return cls.loadPopplerDocument(doc, renderer) if doc else ()

    def mutex(self):
        """No two pages of same Poppler document are rendered at the same time."""
        return self.document

    def group(self):
        """Reimplemented to return the Poppler document our page displays a page from."""
        return self.document

    def ident(self):
        """Reimplemented to return the page number of this page."""
        return self.pageNumber

    def text(self, rect):
        """Returns text inside rectangle."""
        rect = self.mapFromPage(self.pageWidth, self.pageHeight).rect(rect)
        with locking.lock(self.document):
            page = self.document.page(self.pageNumber)
            return page.text(rect)

    def links(self):
        """Reimplemented to use a different caching mechanism."""
        document, pageNumber = self.document, self.pageNumber
        try:
            return _linkscache[document][pageNumber]
        except KeyError:
            with locking.lock(document):
                links = link.Links(map(Link, document.page(pageNumber).links()))
            _linkscache.setdefault(document, {})[pageNumber] = links
            return links


class PopplerDocument(document.SingleSourceDocument):
    """A lazily loaded Poppler (PDF) document."""
    pageClass = PopplerPage

    def __init__(self, source=None, renderer=None):
        super().__init__(source, renderer)
        self._document = None

    def invalidate(self):
        """Reimplemented to clear the Poppler Document reference."""
        super().invalidate()
        self._document = None

    def createPages(self):
        doc = self.document()
        if doc:
            return self.pageClass.loadPopplerDocument(doc, self.renderer)
        return ()

    def document(self):
        """Return the Poppler Document object.

        Returns None if no source was yet set, and False if loading failed.

        """
        if self._document is None:
            source = self.source()
            if source:
                self._document = load(source) or False
        return self._document


class PopplerRenderer(render.AbstractRenderer):
    if popplerqt5:
        renderBackend = popplerqt5.Poppler.Document.SplashBackend
        printRenderBackend = popplerqt5.Poppler.Document.SplashBackend
    else:
        renderBackend = printRenderBackend = 0

    oversampleThreshold = 96

    def render(self, page, key, tile, paperColor=None):
        """Generate an image for the Page referred to by key."""
        if paperColor is None:
            paperColor = page.paperColor or self.paperColor

        doc = page.document
        num = page.pageNumber
        s = page.pageSize()
        if key.rotation & 1:
            s.transpose()

        xres = 72.0 * key.width / s.width()
        yres = 72.0 * key.height / s.height()
        multiplier = 2 if xres < self.oversampleThreshold else 1
        image = self.render_poppler_image(doc, num,
            xres * multiplier, yres * multiplier,
            tile.x * multiplier, tile.y * multiplier, tile.w * multiplier, tile.h * multiplier,
            key.rotation, paperColor)
        if multiplier == 2:
            image = image.scaledToWidth(tile.w, Qt.SmoothTransformation)
        image.setDotsPerMeterX(int(xres * 39.37))
        image.setDotsPerMeterY(int(yres * 39.37))
        return image

    def setRenderHints(self, doc):
        """Set the poppler render hints we want to set."""
        if self.antialiasing:
            doc.setRenderHint(popplerqt5.Poppler.Document.Antialiasing)
            doc.setRenderHint(popplerqt5.Poppler.Document.TextAntialiasing)

    @contextlib.contextmanager
    def setup(self, doc, backend=None, paperColor=None):
        """Use the poppler document in context, properly configured and locked."""
        with locking.lock(doc):
            if backend is not None:
                oldbackend = doc.renderBackend()
                doc.setRenderBackend(backend)
            oldhints = int(doc.renderHints())
            doc.setRenderHint(oldhints, False)
            self.setRenderHints(doc)
            if paperColor is not None:
                oldcolor = doc.paperColor()
                doc.setPaperColor(paperColor)
            try:
                yield
            finally:
                if backend is not None:
                    doc.setRenderBackend(oldbackend)
                doc.setRenderHint(int(doc.renderHints()), False)
                doc.setRenderHint(oldhints)
                if paperColor is not None:
                    doc.setPaperColor(oldcolor)

    def render_poppler_image(self, doc, pageNum,
                                   xres=72.0, yres=72.0,
                                   x=-1, y=-1, w=-1, h=-1, rotate=Rotate_0,
                                   paperColor=None):
        """Render an image, almost like calling page.renderToImage().

        The document is properly locked during rendering and render options
        are set.

        """
        with self.setup(doc, self.renderBackend, paperColor):
            return doc.page(pageNum).renderToImage(xres, yres, x, y, w, h, rotate)

    def draw(self, page, painter, key, tile, paperColor=None):
        """Draw a tile on the painter.

        The painter is already at the right position and rotation.
        For the Poppler page and renderer, draw() is only used for printing.
        (See AbstractPage.print().)

        """
        source = self.map(key, page.pageRect()).mapRect(QRectF(*tile)).toRect()   # rounded
        target = QRectF(0, 0, tile.w, tile.h)
        if key.rotation & 1:
            target.setSize(target.size().transposed())

        doc = page.document
        p = doc.page(page.pageNumber)

        with self.setup(doc, self.printRenderBackend, paperColor):
            if self.printRenderBackend == popplerqt5.Poppler.Document.ArthurBackend:
                # Poppler's Arthur backend removes the current transform from
                # the painter (it sets a default CTM, instead of combining it
                # with the current transform). We let Poppler draw on a QPicture,
                # and draw that on our painter.
                pic = QPicture()
                p.renderToPainter(QPainter(pic), page.dpi, page.dpi, source.x(), source.y(), source.width(), source.height())
                # our resolution could be different, scale accordingly
                painter.save()
                painter.scale(pic.logicalDpiX() / painter.device().logicalDpiX(),
                              pic.logicalDpiY() / painter.device().logicalDpiY())
                pic.play(painter)
                painter.restore()
            else:
                # Make an image exactly in the printer's resolution
                m = painter.transform()
                r = m.mapRect(source)       # see where the source ends up
                w, h = r.width(), r.height()
                if m.m11() == 0:
                    w, h = h, w     # swap if rotation & 1  :-)
                # now we know the scale from our dpi to the paintdevice's logicalDpi!
                hscale = w / source.width()
                vscale = h / source.height()
                s = QTransform().scale(hscale, vscale).mapRect(source)
                dpiX = page.dpi * hscale
                dpiY = page.dpi * vscale
                img = p.renderToImage(dpiX, dpiY, s.x(), s.y(), s.width(), s.height())
                painter.drawImage(target, img, QRectF(img.rect()))


def load(source):
    """Load a Poppler document.

    Source may be:
        - a Poppler document, which is then simply returned :-)
        - a filename
        - q QByteArray instance.

    Returns None if popplerqt5 is not available or the document could not be
    loaded.

    """
    if popplerqt5:
        if isinstance(source, popplerqt5.Poppler.Document):
            return source
        elif isinstance(source, str):
            return popplerqt5.Poppler.Document.load(source)
        else:
            return popplerqt5.Poppler.Document.loadFromData(source)



# install a default renderer, so PopplerPage can be used directly
PopplerPage.renderer = PopplerRenderer()






# -*- coding: utf-8 -*-
#
# This file is part of the qpageview package.
#
# Copyright (c) 2019 - 2019 by Wilbert Berendsen
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 2
# of the License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
# See http://www.gnu.org/licenses/ for more information.

"""
ViewActions provides QActions to control a View.
"""


import weakref

from PyQt5.QtCore import pyqtSignal, QObject, Qt
from PyQt5.QtGui import QKeySequence
from PyQt5.QtWidgets import (
    QAction, QActionGroup, QApplication, QComboBox, QLabel, QSpinBox,
    QWidgetAction)

from . import util
from .constants import *


class ViewActions(QObject):
    """ViewActions provides QActions to control a View.

    Use setView() to connect the actions with a View. If no View is connected,
    and an action is used; the viewRequested signal is emitted. You can connect
    this signal and call setView() in the called slot; the action is then
    performed on the View.

    The attribute `smartLayoutOrientationEnabled` (defaulting to True) enables
    some intuitive behaviour: if set to True, for layout modes that do not make
    sense in horizontal mode the orientation is automatically set to Vertical;
    and when the user chooses Horizontal orientation in such modes, the layout
    mode is set to "single".

    """

    smartLayoutOrientationEnabled = True

    viewRequested = pyqtSignal()

    def __init__(self, *args, **kwargs):
        """Create the actions.

        Does not yet connect anything, use setView() for that.

        """
        super().__init__(*args, **kwargs)
        self._view = lambda: None
        self.createActions()
        self.connectActions()
        self.setActionTexts()
        self.setActionIcons()
        self.setActionShortcuts()

    def setView(self, view):
        """Connects all the actions to the View.

        Use None to set no view. If a view was previously set, all connections
        are removed from that View.

        """
        old = self._view()
        if old == view:
            return
        if old:
            old.viewModeChanged.disconnect(self.updateViewModeActions)
            old.zoomFactorChanged.disconnect(self.updateZoomActions)
            old.pageLayoutModeChanged.disconnect(self.updatePageLayoutModeActions)
            old.orientationChanged.disconnect(self.updateActions)
            old.continuousModeChanged.disconnect(self.updateActions)
            old.currentPageNumberChanged.disconnect(self.updatePagerActions)
            old.pageCountChanged.disconnect(self.updatePagerActions)

        if view:
            view.viewModeChanged.connect(self.updateViewModeActions)
            view.zoomFactorChanged.connect(self.updateZoomActions)
            view.pageLayoutModeChanged.connect(self.updatePageLayoutModeActions)
            view.orientationChanged.connect(self.updateActions)
            view.continuousModeChanged.connect(self.updateActions)
            view.currentPageNumberChanged.connect(self.updatePagerActions)
            view.pageCountChanged.connect(self.updatePagerActions)
            self._view = weakref.ref(view)
            self.updateActions()
            self.updateViewModeActions(view.viewMode())
            self.updateZoomActions(view.zoomFactor())
            self.updatePagerActions()
        else:
            self._view = lambda: None

    def view(self):
        """Return the View.

        If no View is set, viewRequested is emitted. You can connect to this
        signal to create a View, and call setView() to use it to perform
        the requested action.

        """
        view = self._view()
        if not view:
            self.viewRequested.emit()
        return self._view()

    @staticmethod
    def names():
        """Return a tuple of all the names of the actions we support."""
        return (
            'print',
            'fit_width',
            'fit_height',
            'fit_both',
            'zoom_natural',
            'zoom_original'
            'zoom_in',
            'zoom_out',
            'zoomer',
            'rotate_left',
            'rotate_right',
            'layout_single',
            'layout_double_right',
            'layout_double_left',
            'layout_raster',
            'vertical',
            'horizontal',
            'continuous',
            'reload',
            'previous_page',
            'next_page',
            'pager',
            'magnifier',
        )

    def createActions(self):
        """Creates the actions; called by __init__()."""
        self.print = QAction(self)

        self._viewMode = QActionGroup(self)
        self.fit_width = QAction(self._viewMode, checkable=True)
        self.fit_height = QAction(self._viewMode, checkable=True)
        self.fit_both = QAction(self._viewMode, checkable=True)

        self.zoom_natural = QAction(self)
        self.zoom_original = QAction(self)
        self.zoom_in = QAction(self)
        self.zoom_out = QAction(self)
        self.zoomer = ZoomerAction(self)

        self.rotate_left = QAction(self)
        self.rotate_right = QAction(self)

        self._pageLayoutMode = QActionGroup(self)
        self.layout_single = QAction(self._pageLayoutMode, checkable=True)
        self.layout_double_right = QAction(self._pageLayoutMode, checkable=True)
        self.layout_double_left = QAction(self._pageLayoutMode, checkable=True)
        self.layout_raster = QAction(self._pageLayoutMode, checkable=True)

        self._orientation = QActionGroup(self)
        self.vertical = QAction(self._orientation, checkable=True)
        self.horizontal = QAction(self._orientation, checkable=True)

        self.continuous = QAction(self, checkable=True)
        self.reload = QAction(self)

        self.previous_page = QAction(self)
        self.next_page = QAction(self)
        self.pager = PagerAction(self)

        self.magnifier = QAction(self, checkable=True)

    def updateFromProperties(self, properties):
        """Set the actions to the state stored in the given ViewProperties."""
        if properties.pageLayoutMode is not None:
            self.updatePageLayoutModeActions(properties.pageLayoutMode)
        if properties.orientation is not None:
            self.vertical.setChecked(properties.orientation == Vertical)
            self.horizontal.setChecked(properties.orientation == Horizontal)
        if properties.continuousMode is not None:
            self.continuous.setChecked(properties.continuousMode)
        if properties.zoomFactor is not None:
            self.updateZoomActions(properties.zoomFactor)
        if properties.viewMode is not None:
            self.updateViewModeActions(properties.viewMode)

    def connectActions(self):
        """Connect our actions with our methods. Called by __init__()."""
        self.print.triggered.connect(self.slotPrint)
        self._viewMode.triggered.connect(self.slotViewMode)
        self.zoom_natural.triggered.connect(self.slotZoomNatural)
        self.zoom_original.triggered.connect(self.slotZoomOriginal)
        self.zoom_in.triggered.connect(self.slotZoomIn)
        self.zoom_out.triggered.connect(self.slotZoomOut)
        self.zoomer.zoomFactorChanged.connect(self.slotZoomFactor)
        self.zoomer.viewModeChanged.connect(self.slotZoomViewMode)
        self.rotate_left.triggered.connect(self.slotRotateLeft)
        self.rotate_right.triggered.connect(self.slotRotateRight)
        self._pageLayoutMode.triggered.connect(self.slotPageLayoutMode)
        self._orientation.triggered.connect(self.slotOrientation)
        self.continuous.triggered.connect(self.slotContinuousMode)
        self.reload.triggered.connect(self.slotReload)
        self.previous_page.triggered.connect(self.slotPreviousPage)
        self.next_page.triggered.connect(self.slotNextPage)
        self.pager.currentPageNumberChanged.connect(self.slotSetPageNumber)
        self.magnifier.triggered.connect(self.slotMagnifier)

    def updateActions(self):
        """Update the state of the actions not handled in the other update methods."""
        view = self.view()
        if not view:
            return
        self.print.setEnabled(view.pageCount() > 0)
        self.vertical.setChecked(view.orientation() == Vertical)
        self.horizontal.setChecked(view.orientation() == Horizontal)
        self.continuous.setChecked(view.continuousMode())
        self.magnifier.setEnabled(bool(view.magnifier()))
        self.magnifier.setChecked(bool(view.magnifier() and view.magnifier().isVisible()))

    def updatePageLayoutModeActions(self, mode):
        """Update the state of the layout mode actions."""
        self.layout_single.setChecked(mode == "single")
        self.layout_double_left.setChecked(mode == "double_left")
        self.layout_double_right.setChecked(mode == "double_right")
        self.layout_raster.setChecked(mode == "raster")

    def updateViewModeActions(self, mode):
        """Update the state of view mode related actions."""
        self.fit_width.setChecked(mode == FitWidth)
        self.fit_height.setChecked(mode == FitHeight)
        self.fit_both.setChecked(mode == FitBoth)
        self.zoomer.setViewMode(mode)

    def updateZoomActions(self, factor):
        """Update the state of zoom related actions."""
        self.zoomer.setZoomFactor(factor)

    def updatePagerActions(self):
        """Update the state of paging-related actions."""
        view = self.view()
        if not view:
            return
        self.pager.setPageCount(view.pageCount())
        self.pager.updateCurrentPageNumber(view.currentPageNumber())
        self.pager.setEnabled(view.pageCount() > 0)
        self.previous_page.setEnabled(view.currentPageNumber() > 1)
        self.next_page.setEnabled(view.currentPageNumber() < view.pageCount())

    def setActionTexts(self, _=None):
        """Set a default text to all the actions, you may override or translate them.

        You may also set tooltip or whatsthis text in this method.

        """
        if _ is None: _ = lambda t: t
        self.print.setText(_("&Print..."))
        self.fit_width.setText(_("Fit &Width"))
        self.fit_height.setText(_("Fit &Height"))
        self.fit_both.setText(_("Fit &Page"))
        self.zoom_natural.setText(_("&Natural Size"))
        self.zoom_original.setText(_("Original &Size"))
        self.zoom_in.setText(_("Zoom &In"))
        self.zoom_out.setText(_("Zoom &Out"))
        self.zoomer.setViewModes((
            # L10N: "Width" as in "Fit Width" (display in zoom menu)
            (FitWidth, _("Width")),
            # L10N: "Height" as in "Fit Height" (display in zoom menu)
            (FitHeight, _("Height")),
            # L10N: "Page" as in "Fit Page" (display in zoom menu)
            (FitBoth, _("Page")),
        ))
        self.rotate_left.setText(_("Rotate &Left"))
        self.rotate_right.setText(_("Rotate &Right"))
        self.layout_single.setText(_("Single Pages"))
        self.layout_double_right.setText(_("Two Pages (first page right)"))
        self.layout_double_left.setText(_("Two Pages (first page left)"))
        self.layout_raster.setText(_("Raster"))
        self.vertical.setText(_("Vertical"))
        self.horizontal.setText(_("Horizontal"))
        self.continuous.setText(_("&Continuous"))
        self.reload.setText(_("Re&load View"))
        self.previous_page.setText(_("Previous Page"))
        self.previous_page.setIconText(_("Previous"))
        self.next_page.setText(_("Next Page"))
        self.next_page.setIconText(_("Next"))
        self.magnifier.setText(_("Magnifier"))

    def setActionIcons(self):
        """Implement this method to set icons to the actions."""
        pass

    def setActionShortcuts(self):
        """Implement this method to set keyboard shortcuts to the actions."""
        self.print.setShortcuts(QKeySequence.Print)
        self.zoom_in.setShortcuts(QKeySequence.ZoomIn)
        self.zoom_out.setShortcuts(QKeySequence.ZoomOut)
        self.reload.setShortcut(QKeySequence(Qt.Key_F5))

    def slotPrint(self):
        view = self.view()
        if view:
            view.print()

    def slotViewMode(self, action):
        view = self.view()
        if view:
            viewMode = FitWidth if action == self.fit_width else \
                       FitHeight if action == self.fit_height else \
                       FitBoth
            view.setViewMode(viewMode)

    def slotZoomNatural(self):
        view = self.view()
        if view:
            view.zoomNaturalSize()

    def slotZoomOriginal(self):
        view = self.view()
        if view:
            view.setZoomFactor(1.0)

    def slotZoomIn(self):
        view = self.view()
        if view:
            view.zoomIn()

    def slotZoomOut(self):
        view = self.view()
        if view:
            view.zoomOut()

    def slotZoomViewMode(self, mode):
        view = self.view()
        if view:
            view.setViewMode(mode)

    def slotZoomFactor(self, factor):
        view = self.view()
        if view:
            view.setZoomFactor(factor)

    def slotRotateLeft(self):
        view = self.view()
        if view:
            view.rotateLeft()

    def slotRotateRight(self):
        view = self.view()
        if view:
            view.rotateRight()

    def slotPageLayoutMode(self, action):
        view = self.view()
        if view:
            mode = "single" if action == self.layout_single else \
                   "double_left" if action == self.layout_double_left else \
                   "double_right" if action == self.layout_double_right else \
                   "raster"
            view.setPageLayoutMode(mode)

            if self.smartLayoutOrientationEnabled:
                if mode in ("double_left", "double_right"):
                    view.setOrientation(Vertical)

    def slotOrientation(self, action):
        view = self.view()
        if view:
            orientation = Vertical if action == self.vertical else Horizontal
            view.setOrientation(orientation)

            if self.smartLayoutOrientationEnabled:
                if orientation == Horizontal and \
                        view.pageLayoutMode() in ("double_left", "double_right"):
                    view.setPageLayoutMode("single")

    def slotContinuousMode(self):
        view = self.view()
        if view:
            view.setContinuousMode(self.continuous.isChecked())

    def slotReload(self):
        view = self.view()
        if view:
            view.reload()

    def slotPreviousPage(self):
        view = self.view()
        if view:
            view.gotoPreviousPage()

    def slotNextPage(self):
        view = self.view()
        if view:
            view.gotoNextPage()

    def slotSetPageNumber(self, num):
        view = self.view()
        if view:
            view.setCurrentPageNumber(num)

    def slotMagnifier(self):
        view = self._view() # do not trigger creation
        if view and view.magnifier():
            view.magnifier().setVisible(self.magnifier.isChecked())


class PagerAction(QWidgetAction):
    """PagerAction shows a spinbox widget with the current page number.

    When the current page number is changed (by the user or by calling
    setCurrentPageNumber()) the signal currentPageNumberChanged() is emitted
    with the new current page number.

    You can use the instance or class attributes buttonSymbols, focusPolicy and
    the displayFormat() method to influence behaviour and appearance of the
    spinbox widget(s) that is/are created when this action is added to a
    toolbar.

    The displayFormat string should contain the text "{num}". You can also
    include the string "{total}", so the page count is displayed as well.

    """

    currentPageNumberChanged = pyqtSignal(int)

    buttonSymbols = QSpinBox.NoButtons
    focusPolicy = Qt.ClickFocus

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self._currentPage = 0
        self._pageCount = 0
        self._displayFormat = "{num} of {total}"

    def createWidget(self, parent):
        w = QSpinBox(parent, buttonSymbols=self.buttonSymbols)
        w.setFocusPolicy(self.focusPolicy)
        self._adjustSpinBox(w)
        if self._currentPage:
            w.setValue(self._currentPage)
        w.valueChanged[int].connect(self.setCurrentPageNumber)
        return w

    def setButtonSymbols(self, buttonSymbols):
        """Set the ``buttonSymbols`` property, and update already existing widgets."""
        self.buttonSymbols = buttonSymbols
        for w in self.createdWidgets():
            w.setButtonSymbols(buttonSymbols)

    def displayFormat(self):
        """Return the currently active display format string."""
        return self._displayFormat

    def setDisplayFormat(self, displayFormat):
        """Set the display format string to use.

        The default is "{num} of {total}".

        """
        assert "{num}" in displayFormat
        if displayFormat != self._displayFormat:
            self._displayFormat = displayFormat
            self._updateDisplay()

    def pageCount(self):
        """Return the currently set page count."""
        return self._pageCount

    def setPageCount(self, pageCount):
        """Set the page count."""
        if pageCount != self._pageCount:
            self._pageCount = pageCount
            if pageCount:
                self._currentPage = max(1, min(self._currentPage, pageCount))
            else:
                self._currentPage = 0
            self._updateDisplay()

    def currentPageNumber(self):
        """Return the current page number."""
        return self._currentPage

    def setCurrentPageNumber(self, num):
        """Set our current page number."""
        if num and num != self._currentPage:
            self.updateCurrentPageNumber(num)
            self.currentPageNumberChanged.emit(num)

    def updateCurrentPageNumber(self, num):
        """Set our current page number, but without emitting the signal."""
        if num and num != self._currentPage:
            self._currentPage = num
            for w in self.createdWidgets():
                w.setValue(num)
                w.lineEdit().deselect()

    def _adjustSpinBox(self, widget):
        """Update the display of the individual spinbox."""
        if self._pageCount:
            if "{num}" in self._displayFormat:
                prefix, suffix = self._displayFormat.split('{num}', 1)
            else:
                prefix, suffix = "", ""
            widget.setSpecialValueText("")
            widget.setRange(1, self._pageCount)
            widget.setSuffix(suffix.format(total=self._pageCount))
            widget.setPrefix(prefix.format(total=self._pageCount))
        else:
            widget.setSpecialValueText(" ")
            widget.setRange(0, 0)
            widget.setSuffix("")
            widget.setPrefix("")
            widget.clear()

    def _updateDisplay(self):
        """Update the display in the pager.

        This is called when the page count or the display format string
        is changed.

        """
        for w in self.createdWidgets():
            self._adjustSpinBox(w)


class ZoomerAction(QWidgetAction):
    """ZoomerAction provides a combobox with view modes and zoom factors."""

    zoomFactorChanged = pyqtSignal(float)
    viewModeChanged = pyqtSignal(int)

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self._zoomFactor = 1.0
        self._viewMode = FixedScale
        self._viewModes = (
            (FitWidth, "Width"),
            (FitHeight, "Height"),
            (FitBoth, "Page"),
        )
        self._zoomFactors = (
            0.5,
            0.75,
            1.0,
            1.25,
            1.5,
            2.0,
            3.0,
            8.0,
            24.0,
            64.0,
        )
        self._zoomFormat = "{0:.0%}"

    def viewModes(self):
        """Return the view modes that are displayed in the combobox.

        See setViewModes() for explanation.

        """
        return self._viewModes

    def setViewModes(self, modes):
        """Set the view modes to display on top of the zoom values in the box.

        An iterable of tuples (mode, name) is expected; every mode is a
        viewMode, the name is displayed. By default modes 1, 2 and 3 are
        displayed with the names "Width", "Height", "Page".

        """
        self._viewModes = tuple(modes)
        self._setupComboBoxes()

    def zoomFactors(self):
        """Return the zoom factors that are displayed in the combobox.

        A zoom factor of 100% is represented by a floating point value of 1.0.

        """
        return self._zoomFactors

    def setZoomFactors(self, factors):
        """Set the zoom factors to display in the combobox.

        A zoom factor of 100% is represented by a floating point value of 1.0.

        """
        self._zoomFactors = tuple(factors)
        self._setupComboBoxes()

    def zoomFormat(self):
        """Return the format string used to display zoom factors."""
        return self._zoomFormat

    def setZoomFormat(self, zoomFormat):
        """Set the format string used to display zoom factors."""
        self._zoomFormat = zoomFormat
        self._setupComboBoxes()

    def createWidget(self, parent):
        w = QComboBox(parent)
        w.setSizeAdjustPolicy(QComboBox.AdjustToContents)
        w.setEditable(True)
        w.lineEdit().setReadOnly(True)
        w.setFocusPolicy(Qt.NoFocus)
        self._setupComboBox(w)
        self._adjustComboBox(w)
        w.activated[int].connect(self.setCurrentIndex)
        return w

    def viewMode(self):
        """Return the current view mode."""
        return self._viewMode

    def setViewMode(self, mode):
        """Set the current view mode."""
        if mode != self._viewMode:
            self._viewMode = mode
            self.viewModeChanged.emit(mode)
            self._adjustComboBoxes()

    def zoomFactor(self):
        """Return the current zoom factor."""
        return self._zoomFactor

    def setZoomFactor(self, factor):
        """Set the current zoom factor."""
        if factor != self._zoomFactor:
            self._zoomFactor = factor
            self.zoomFactorChanged.emit(factor)
            self._adjustComboBoxes()

    def setCurrentIndex(self, index):
        """Called when the user chooses an entry in a combobox."""
        viewModeCount = len(self._viewModes)
        if index < viewModeCount:
            self.setViewMode(self._viewModes[index][0])
        else:
            self.setZoomFactor(self._zoomFactors[index - viewModeCount])

    def _setupComboBoxes(self):
        """Update the contents and current setting of all comboboxes.

        Called after setting view modes and zoom values.

        """
        for w in self.createdWidgets():
            with util.signalsBlocked(w):
                self._setupComboBox(w)
                self._adjustComboBox(w)

    def _adjustComboBoxes(self):
        """Adjust the current setting (zoom/viewmode) of all comboboxes.

        Called when current zoom or view mode changes.

        """
        for w in self.createdWidgets():
            with util.signalsBlocked(w):
                self._adjustComboBox(w)

    def _setupComboBox(self, w):
        """Put the entries in the (new) QComboBox widget."""
        w.clear()
        for mode, name in self._viewModes:
            w.addItem(name)
        for v in self._zoomFactors:
            w.addItem(self._zoomFormat.format(v))

    def _adjustComboBox(self, w):
        """Select the current index based on our zoomFactor and view mode."""
        for i, (mode, name) in enumerate(self._viewModes):
            if mode == self._viewMode:
                w.setCurrentIndex(i)
                break
        else:
            if self._zoomFactor in self._zoomFactors:
                i = self._zoomFactors.index(self._zoomFactor) + len(self._viewModes)
                w.setCurrentIndex(i)
            else:
                w.setEditText(self._zoomFormat.format(self._zoomFactor))





# -*- coding: utf-8 -*-
#
# This file is part of the qpageview package.
#
# Copyright (c) 2010 - 2019 by Wilbert Berendsen
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 2
# of the License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
# See http://www.gnu.org/licenses/ for more information.


"""
Manages lists of rectangular objects and quickly finds them.
"""

import bisect
import operator


Left   = 0
Top    = 1
Right  = 2
Bottom = 3


class Rectangles:
    """
    Manages a list of rectangular objects and quickly finds objects at
    some point, in some rectangle or intersecting some rectangle.

    The implementation uses four lists of the objects sorted on either
    coordinate, so retrieval is fast.

    Bulk adding is done in the constructor or via the bulk_add() method (which
    clears the indexes, that are recreated on first search).  Single objects
    can be added and deleted, keeping the indexes, but that's slower.

    You should inherit from this class and implement the method get_coords(obj)
    to get the rectangle of the object (x, y, x2, y2). These are requested only
    once. x should be < x2 and y should be < y2.

    """
    def __init__(self, objects=None):
        """Initializes the Rectangles object.

        objects should, if given, be an iterable of rectangular objects, and
        bulk_add() is called on those objects.

        """
        self._items = {} # maps object to the result of func(object)
        self._index = {} # maps side to indices, objects (index=coordinate of that side)
        if objects:
            self.bulk_add(objects)

    def get_coords(self, obj):
        """You should implement this method.

        The result should be a four-tuple with the coordinates of the rectangle
        the object represents (x, y, x2, y2). These are requested only once.
        x should be < x2 and y should be < y2.

        """
        return (0, 0, 0, 0)

    def add(self, obj):
        """Adds an object to our list. Keeps the index intact."""
        if obj in self._items:
            return
        self._items[obj] = coords = self.get_coords(obj)
        for side, (indices, objects) in self._index.items():
            i = bisect.bisect_left(indices, coords[side])
            indices.insert(i, coords[side])
            objects.insert(i, obj)

    def bulk_add(self, objects):
        """Adds many new items to the index using the function given in the constructor.

        After this, the index is cleared and recreated on the first search operation.

        """
        self._items.update((obj, self.get_coords(obj)) for obj in objects)
        self._index.clear()

    def remove(self, obj):
        """Removes an object from our list. Keeps the index intact."""
        del self._items[obj]
        for indices, objects in self._index.values():
            i = objects.index(obj)
            del objects[i]
            del indices[i]

    def clear(self):
        """Empties the list of items."""
        self._items.clear()
        self._index.clear()

    def at(self, x, y):
        """Returns a set() of objects that are touched by the given point."""
        return self._test(
            (self._smaller, Top, y),
            (self._larger, Bottom, y),
            (self._smaller, Left, x),
            (self._larger, Right, x))

    def inside(self, left, top, right, bottom):
        """Returns a set() of objects that are fully in the given rectangle."""
        return self._test(
            (self._larger, Top, top),
            (self._smaller, Bottom, bottom),
            (self._larger, Left, left),
            (self._smaller, Right, right))

    def intersecting(self, left, top, right, bottom):
        """Returns a set() of objects intersecting the given rectangle."""
        return self._test(
            (self._smaller, Top, bottom),
            (self._larger, Bottom, top),
            (self._smaller, Left, right),
            (self._larger, Right, left))

    def width(self, obj):
        """Return the width of the specified object.

        This can be used for sorting a set returned by at(), inside() or
        intersecting(). For example::

            for r in sorted(rects.at(10, 20), key=rects.width):
                # ...

        """
        coords = self._items[obj]
        return coords[Right] - coords[Left]

    def height(self, obj):
        """Return the height of the specified object. See also width()."""
        coords = self._items[obj]
        return coords[Bottom] - coords[Top]

    def closest(self, obj, side):
        """Returns the object closest to the given one, going to the given side."""
        coords = self._items[obj]
        pos = coords[side^2]
        lat = (coords[side^1|2] - coords[side^1&2]) / 2.0
        direction = -1 if side < Right else 1
        indices, objects = self._sorted(side^2)
        i = objects.index(obj)
        mindist = indices[-1]
        result = []
        for other in objects[i+direction::direction]:
            coords = self._items[other]
            pos1 = coords[side^2]
            d = abs(pos1 - pos)
            if d > mindist:
                break
            lat1 = (coords[side^1|2] - coords[side^1&2]) / 2.0
            dlat = abs(lat1 - lat)
            if dlat < d:
                dist = dlat + d  # manhattan dist
                result.append((other, dist))
                mindist = min(mindist, dist)
        if result:
            result.sort(key=lambda r: r[1])
            return result[0][0]

    def nearest(self, x, y):
        """Return the object with the shortest distance to the point x, y.

        The point (x, y) is outside the object. Use at() to get objects that
        touch the point (x, y). If there are no objects, None is returned.

        """
        i = self._items

        left = self._larger(Left, x)            # closest one is first
        right = self._smaller(Right, x)         # closest one is last
        top = self._larger(Top, y)              # closest one is first
        bottom = self._smaller(Bottom, y)       # closest one is last

        result = []

        # first find adjacent rectangles. For each side, as soon as one is
        # found, don't look further for that side. Only save rectangles that are
        # closer but not adjacent, they could be closer on another side.
        left_over = 0
        for o in left:
            if o not in top and o not in bottom:
                result.append((i[o][Left] - x, o))
                break
            left_over += 1
        top_over = 0
        for o in top:
            if o not in left and o not in right:
                result.append((i[o][Top] - y, o))
                break
            top_over += 1
        right_over = 0
        for o in right[::-1]:
            if o not in top and o not in bottom:
                result.append((x - i[o][Right], o))
                break
            right_over -= 1
        bottom_over = 0
        for o in bottom[::-1]:
            if o not in left and o not in right:
                result.append((y - i[o][Bottom], o))
                break
            bottom_over -= 1
        # at most 4 rectangles are found, the closest one on each edge.
        # Now look for rectangles that could be closer at the corner.
        if left_over and top_over:
            for o in set(left[:left_over]).intersection(top[:top_over]):
                result.append((i[o][Left] - x + i[o][Top] - y, o))
        if top_over and right_over:
            for o in set(top[:top_over]).intersection(right[right_over:]):
                result.append((i[o][Top] - y + x - i[o][Right], o))
        if left_over and bottom_over:
            for o in set(left[:left_over]).intersection(bottom[bottom_over:]):
                result.append((i[o][Left] - x + y - i[o][Bottom], o))
        if bottom_over and right_over:
            for o in set(bottom[bottom_over:]).intersection(right[right_over:]):
                result.append((y - i[o][Bottom] + x - i[o][Right], o))

        if result:
            return min(result, key=operator.itemgetter(0))[1]

    def __len__(self):
        """Return the number of objects."""
        return len(self._items)

    def __contains__(self, obj):
        """Return True if the object is managed by us."""
        return obj in self._items

    def __bool__(self):
        """Always return True."""
        return True

    def __iter__(self):
        """Iterate over the objects in undefined order."""
        return iter(self._items)

    # private helper methods
    def _test(self, *tests):
        """Performs tests and returns objects that fulfill all of them.

        Every test should be a three tuple(method, side, value).
        Method is either self._smaller or self._larger.
        Returns a (possibly empty) set.

        """
        meth, side, value = tests[0]
        result = set(meth(side, value))
        if result:
            for meth, side, value in tests[1:]:
                result &= set(meth(side, value))
                if not result:
                    break
        return result

    def _smaller(self, side, value):
        """Returns objects for side below value."""
        indices, objects = self._sorted(side)
        i = bisect.bisect_right(indices, value)
        return objects[:i]

    def _larger(self, side, value):
        """Returns objects for side above value."""
        indices, objects = self._sorted(side)
        i = bisect.bisect_left(indices, value)
        return objects[i:]

    def _sorted(self, side):
        """Returns a two-tuple (indices, objects) sorted on index for the given side."""
        try:
            return self._index[side]
        except KeyError:
            if self._items:
                objects = [(coords[side], obj) for obj, coords in self._items.items()]
                objects.sort(key=operator.itemgetter(0))
                result = tuple(map(list, zip(*objects)))
            else:
                result = [], []
            self._index[side] = result
            return result





# -*- coding: utf-8 -*-
#
# This file is part of the qpageview package.
#
# Copyright (c) 2019 - 2019 by Wilbert Berendsen
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 2
# of the License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
# See http://www.gnu.org/licenses/ for more information.

"""
Export Pages to different file formats.
"""

import os

from PyQt5.QtCore import QBuffer, QIODevice, QMimeData, QPoint, QSizeF, Qt, QUrl
from PyQt5.QtGui import QDrag, QGuiApplication, QImage, QPageSize, QPdfWriter

from . import poppler
from . import util


class AbstractExporter:
    """Base class to export a rectangular area of a Page to a file.

    Specialized subclasses implement each format.

    You instantiate a subclass with a Page and a rectangle. The rectangle may
    be None, to specify the full page. After instantiation, you can set
    attributes to configure the export. The following attributes are supported::

        resolution = 300
        autocrop = False
        oversample = 1
        grayscale = False
        paperColor = None
        forceVector = True  # force the render backend to be Arthur for
                            # exporting PDF pages to vector-based formats

    After setting the attributes, you call one or more of save(), copyData(),
    copyFile(), mimeData() or tempFileMimeData(), which will trigger the export
    because they internally call data(), which caches its return value until
    setPage() is called again.

    Not all exporters support all attributes, the supportXXX attributes specify
    whether an attribute is supported or not.

    """
    # user settings:
    resolution = 300
    antialiasing = True
    autocrop = False
    oversample = 1
    grayscale = False
    paperColor = None
    forceVector = True  # force the render backend to be Arthur for PDF pages

    # properties of exporter:
    wantsVector = True
    supportsResolution = True
    supportsAntialiasing = True
    supportsAutocrop = True
    supportsOversample = True
    supportsGrayscale = True
    supportsPaperColor = True

    mimeType = "application/octet-stream"
    filename = ""
    defaultBasename = "document"
    defaultExt = ""

    def __init__(self, page, rect=None):
        self.setPage(page, rect)

    def setPage(self, page, rect=None):
        self._page = page.copy()
        if self._page.renderer:
            self._page.renderer = page.renderer.copy()
        self._rect = rect
        self._result = None   # where the exported object is stored
        self._tempFile = None
        self._autoCropRect = None
        self._document = None
        self._pixmap = None

    def page(self):
        """Return our page, setting the renderer to our preferences."""
        p = self._page.copy()
        p.paperColor = self.paperColor
        if self._page.renderer:
            p.renderer = self._page.renderer.copy()
            p.renderer.paperColor = self.paperColor
            p.renderer.antialiasing = self.antialiasing
            if self.forceVector and self.wantsVector and \
                    isinstance(p, poppler.PopplerPage) and poppler.popplerqt5:
                p.renderer.printRenderBackend = \
                    poppler.popplerqt5.Poppler.Document.ArthurBackend
        return p

    def autoCroppedRect(self):
        """Return the rect, autocropped if desired."""
        if not self.autocrop:
            return self._rect
        if self._autoCropRect is None:
            p = self._page
            dpiX = p.width / p.defaultSize().width() * p.dpi
            dpiY = p.height / p.defaultSize().height() * p.dpi
            image = p.image(self._rect, dpiX, dpiY)
            rect = util.autoCropRect(image)
            # add one pixel to prevent loosing small joins or curves etc
            rect = image.rect() & rect.adjusted(-1, -1, 1, 1)
            if self._rect is not None:
                rect.translate(self._rect.topLeft())
            self._autoCropRect = rect
        return self._autoCropRect

    def export(self):
        """Perform the export, based on the settings, and return the exported data object."""

    def successful(self):
        """Return True when export was successful."""
        return self.data() is not None

    def data(self):
        """Return the export result, assuming it is binary data of the exported file."""
        if self._result is None:
            self._result = self.export()
        return self._result

    def document(self):
        """Return a one-page Document to display the image to export.

        Internally calls createDocument(), and caches the result, setting the
        papercolor to the papercolor attribute if the exporter supports
        papercolor.

        """
        if self._document is None:
            doc = self._document = self.createDocument()
            if self.paperColor and self.paperColor.isValid():
                for p in doc.pages():
                    p.paperColor = self.paperColor
        return self._document

    def createDocument(self):
        """Create and return a one-page Document to display the image to export."""

    def renderer(self):
        """Return a renderer for the document(). By default, None is returned."""
        return None

    def copyData(self):
        """Copy the QMimeData() to the clipboard."""
        QGuiApplication.clipboard().setMimeData(self.mimeData())

    def mimeData(self):
        """Return a QMimeData() object representing the exported data."""
        data = QMimeData()
        data.setData(self.mimeType, self.data())
        return data

    def save(self, filename):
        """Save the exported image to a file."""
        with open(filename, "wb") as f:
            f.write(self.data())

    def suggestedFilename(self):
        """Return a suggested file name for the file to export.

        The name is based on the filename (if set) and also contains the
        directory path. But the name will never be the same as the filename
        set in the filename attribute.

        """
        if self.filename:
            base = os.path.splitext(self.filename)[0]
            name = base + self.defaultExt
            if name == self.filename:
                name = base + "-export" + self.defaultExt
        else:
            name = self.defaultBasename + self.defaultExt
        return name

    def tempFilename(self):
        """Save data() to a tempfile and returns the filename."""
        if self._tempFile is None:
            if self.filename:
                basename = os.path.splitext(os.path.basename(self.filename))[0]
            else:
                basename = self.defaultBasename
            d = util.tempdir()
            fname = self._tempFile = os.path.join(d, basename + self.defaultExt)
            self.save(fname)
        return self._tempFile

    def tempFileMimeData(self):
        """Save the exported image to a temp file and return a QMimeData object for the url."""
        data = QMimeData()
        data.setUrls([QUrl.fromLocalFile(self.tempFilename())])
        return data

    def copyFile(self):
        """Save the exported image to a temp file and copy its name to the clipboard."""
        QGuiApplication.clipboard().setMimeData(self.tempFileMimeData())

    def pixmap(self, size=100):
        """Return a small pixmap to use for dragging etc."""
        if self._pixmap is None:
            paperColor = self.paperColor if self.supportsPaperColor else None
            page = self.document().pages()[0]
            self._pixmap = page.pixmap(paperColor=paperColor)
        return self._pixmap

    def drag(self, parent, mimeData):
        """Called by dragFile and dragData. Execs a QDrag on the mime data."""
        d = QDrag(parent)
        d.setMimeData(mimeData)
        d.setPixmap(self.pixmap())
        d.setHotSpot(QPoint(-10, -10))
        return d.exec_(Qt.CopyAction)

    def dragData(self, parent):
        """Start dragging the data. Parent can be any QObject."""
        return self.drag(parent, self.mimeData())

    def dragFile(self, parent):
        """Start dragging the data. Parent can be any QObject."""
        return self.drag(parent, self.tempFileMimeData())


class ImageExporter(AbstractExporter):
    """Export a rectangular area of a Page (or the whole page) to an image."""
    wantsVector = False
    defaultBasename = "image"
    defaultExt = ".png"

    def export(self):
        """Create the QImage representing the exported image."""
        res = self.resolution
        if self.oversample != 1:
            res *= self.oversample
        i = self.page().image(self._rect, res, res, self.paperColor)
        if self.oversample != 1:
            i = i.scaled(i.size() / self.oversample, transformMode=Qt.SmoothTransformation)
        if self.grayscale:
            i = i.convertToFormat(QImage.Format_Grayscale8)
        if self.autocrop:
            i = i.copy(util.autoCropRect(i))
        return i

    def image(self):
        return self.data()

    def createDocument(self):
        from . import image
        return image.ImageDocument([self.image()], self.renderer())

    def copyData(self):
        QGuiApplication.clipboard().setImage(self.image())

    def mimeData(self):
        data = QMimeData()
        data.setImageData(self.image())
        return data

    def save(self, filename):
        if not self.image().save(filename):
            raise OSError("Could not save image")


class SvgExporter(AbstractExporter):
    """Export a rectangular area of a Page (or the whole page) to a SVG file."""
    mimeType = "image/svg"
    supportsGrayscale = False
    supportsOversample = False
    defaultBasename = "image"
    defaultExt = ".svg"

    def export(self):
        rect = self.autoCroppedRect()
        buf = QBuffer()
        buf.open(QBuffer.WriteOnly)
        success = self.page().svg(buf, rect, self.resolution, self.paperColor)
        buf.close()
        if success:
            return buf.data()

    def createDocument(self):
        from . import svg
        return svg.SvgDocument([self.data()], self.renderer())


class PdfExporter(AbstractExporter):
    """Export a rectangular area of a Page (or the whole page) to a PDF file."""
    mimeType = "application/pdf"
    supportsGrayscale = False
    supportsOversample = False
    defaultExt = ".pdf"

    def export(self):
        rect = self.autoCroppedRect()
        buf = QBuffer()
        buf.open(QBuffer.WriteOnly)
        success = self.page().pdf(buf, rect, self.resolution, self.paperColor)
        buf.close()
        if success:
            return buf.data()

    def createDocument(self):
        from . import poppler
        return poppler.PopplerDocument(self.data(), self.renderer())


class EpsExporter(AbstractExporter):
    """Export a rectangular area of a Page (or the whole page) to an EPS file."""
    mimeType = "application/postscript"
    supportsGrayscale = False
    supportsOversample = False
    defaultExt = ".eps"

    def export(self):
        rect = self.autoCroppedRect()
        buf = QBuffer()
        buf.open(QBuffer.WriteOnly)
        success = self.page().eps(buf, rect, self.resolution, self.paperColor)
        buf.close()
        if success:
            return buf.data()

    def createDocument(self):
        from . import poppler
        rect = self.autoCroppedRect()
        buf = QBuffer()
        buf.open(QBuffer.WriteOnly)
        success = self.page().pdf(buf, rect, self.resolution, self.paperColor)
        buf.close()
        return poppler.PopplerDocument(buf.data(), self.renderer())


def pdf(filename, pageList, resolution=72, paperColor=None):
    """Export the pages in pageList to a PDF document.

    filename can be a string or any QIODevice. The pageList is a list of the
    Page objects to export.

    Normally vector graphics are rendered, but in cases where that is not
    possible, the resolution will be used to determine the DPI for the
    generated rendering.

    The computedRotation attribute of the pages is used to determine the
    rotation.

    Make copies of the pages if you run this function in a background thread.

    """
    pdf = QPdfWriter(filename)
    pdf.setCreator("qpageview")
    pdf.setResolution(resolution)

    for n, page in enumerate(pageList):
        # map to the original page
        source = page.pageRect()
        # scale to target size
        w = source.width() * page.scaleX
        h = source.height() * page.scaleY
        if page.computedRotation & 1:
            w, h = h, w
        targetSize = QSizeF(w, h)
        if n:
            pdf.newPage()
        layout = pdf.pageLayout()
        layout.setMode(layout.FullPageMode)
        layout.setPageSize(QPageSize(targetSize * 72.0 / page.dpi, QPageSize.Point))
        pdf.setPageLayout(layout)
        # TODO handle errors?
        page.output(pdf, source, paperColor)





# -*- coding: utf-8 -*-
#
# This file is part of the qpageview package.
#
# Copyright (c) 2016 - 2019 by Wilbert Berendsen
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 2
# of the License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
# See http://www.gnu.org/licenses/ for more information.

"""
Constant values.
"""


# rotation:
Rotate_0   = 0      #: normal
Rotate_90  = 1      #: 90° rotated clockwise
Rotate_180 = 2      #: 180° rotated
Rotate_270 = 3      #: 270° rotated (90° couter clockwise)


# viewModes:
FixedScale = 0      #: the scale is not adjusted to the widget size
FitWidth   = 1      #: scale so that the page's width fits in the widget
FitHeight  = 2      #: scale so that the page's height fits in the widget
FitBoth    = FitHeight | FitWidth   #: fit the whole page


# orientation:
Horizontal = 1      #: arrange the pages in horizontal order
Vertical   = 2      #: arrange the pages in vertical order




# -*- coding: utf-8 -*-
#
# This file is part of the qpageview package.
#
# Copyright (c) 2019 - 2019 by Wilbert Berendsen
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 2
# of the License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
# See http://www.gnu.org/licenses/ for more information.

"""
SelectorViewMixin class, to mixin with View.

Adds the capability to select or unselect Pages.

"""

import contextlib

from PyQt5.QtCore import pyqtSignal, QRect, Qt
from PyQt5.QtGui import QPainter, QKeySequence
from PyQt5.QtWidgets import QStyle, QStyleOptionButton


class SelectorViewMixin:
    """SelectorViewMixin class, to mixin with View.

    Adds the capability to select or unselect Pages.
    Pages are numbered from 1.

    Instance variables:

    ``userChangeSelectionModeEnabled`` = True
        whether the user can change the selectionMode (by longpressing a page
        to enable selectionMode, and pressing ESC to leave selectionMode. (Be
        sure to mix in the :class:`qpageview.util.LongMousePressMixin` class
        when you want to use the long mouse press event.)

    """
    selectionChanged = pyqtSignal()
    selectionModeChanged = pyqtSignal(bool)

    userChangeSelectionModeEnabled = True


    def __init__(self, parent=None, **kwds):
        self._selection = set()
        self._selectionMode = False
        super().__init__(parent, **kwds)

    def selection(self):
        """Return the current list of selected page numbers."""
        return sorted(self._selection)

    @contextlib.contextmanager
    def modifySelection(self):
        """Context manager that allows changing the selection.

        Yields a set, and on exit of the context, stores the modifications and
        emits the selectionChanged() signal. Used internally by all other
        methods.

        """
        old = set(self._selection)
        yield self._selection
        self._checkSelection()
        diff = self._selection ^ old
        if diff:
            self.selectionChanged.emit()
            # repaint if needed
            visible = self.visibleRect()
            if any(self.page(n).geometry() & visible for n in diff):
                self.viewport().update()

    def updatePageLayout(self, lazy=False):
        """Reimplemented to also check the selection."""
        super().updatePageLayout(lazy)
        self._checkSelection()

    def _checkSelection(self):
        """Internal; silently remove page numbers from the selection that do not exist (anymore)."""
        count = self.pageCount()
        nums = [n for n in self._selection if n < 1 or n > count]
        self._selection.difference_update(nums)

    def clearSelection(self):
        """Convenience method to clear the selection."""
        with self.modifySelection() as s:
            s.clear()

    def selectAll(self):
        """Convenience method to select all pages."""
        with self.modifySelection() as s:
            s.update(range(1, self.pageCount() + 1))

    def toggleSelection(self, pageNumber):
        """Toggles the selected state of page number pageNumber."""
        with self.modifySelection() as s:
            count = len(s)
            s.add(pageNumber)
            if count == len(s):
                s.remove(pageNumber)

    def selectionMode(self):
        """Return the current selectionMode (True is enabled, False is disabled)."""
        return self._selectionMode

    def setSelectionMode(self, mode):
        """Switch selection mode on or off (True is enabled, False is disabled)."""
        if self._selectionMode != mode:
            self._selectionMode = mode
            self.selectionModeChanged.emit(mode)
            self.viewport().update() # repaint

    def paintEvent(self, ev):
        super().paintEvent(ev)      # first draw the contents
        if self._selectionMode:
            painter = QPainter(self.viewport())
            for page, rect in self.pagesToPaint(ev.rect(), painter):
                self.drawSelection(page, painter)

    def drawSelection(self, page, painter):
        """Draws the state (selected or not) for the page."""
        option = QStyleOptionButton()
        option.initFrom(self)
        option.rect = QRect(0, 0, QStyle.PM_IndicatorWidth, QStyle.PM_IndicatorHeight)
        pageNum = self.pageLayout().index(page) + 1
        option.state |= QStyle.State_On if pageNum in self._selection else QStyle.State_Off
        scale = None
        # in the unlikely case the checkboxes are larger than the page, scale them down
        if option.rect not in page.rect():
            scale = min(page.width / option.rect.width(), page.height / option.rect.height())
            painter.save()
            painter.scale(scale, scale)
        self.style().drawPrimitive(QStyle.PE_IndicatorCheckBox, option, painter, self)
        if scale is not None:
            painter.restore()

    def mousePressEvent(self, ev):
        """Reimplemented to check if a checkbox was clicked."""
        if self._selectionMode and ev.buttons() == Qt.LeftButton:
            pos = ev.pos() - self.layoutPosition()
            page = self._pageLayout.pageAt(pos)
            if page:
                pageNum = self._pageLayout.index(page) + 1
                pos -= page.pos()
                if pos in QRect(0, 0, QStyle.PM_IndicatorWidth, QStyle.PM_IndicatorHeight):
                    # the indicator has been clicked
                    if ev.modifiers() & Qt.ControlModifier:
                        # CTRL toggles selection of page
                        self.toggleSelection(pageNum)
                    elif self._selection and ev.modifiers() & Qt.ShiftModifier:
                        # Shift extends the selection
                        with self.modifySelection() as s:
                            s.add(pageNum)
                            first, last = min(self._selection), max(self._selection)
                            s.update(range(first, last+1))
                    else:
                        # toggle this one and clear all the others
                        with self.modifySelection() as s:
                            select = pageNum not in s
                            s.clear()
                            if select:
                                s.add(pageNum)
                    return
        super().mousePressEvent(ev)

    def keyPressEvent(self, ev):
        """Clear the selection and switch off selectionmode with ESC."""
        if self._selectionMode:
            if self.userChangeSelectionModeEnabled and ev.key() == Qt.Key_Escape and not ev.modifiers():
                self.clearSelection()
                self.setSelectionMode(False)
                return
            elif ev.matches(QKeySequence.SelectAll):
                self.selectAll()
                return
        super().keyPressEvent(ev)

    def longMousePressEvent(self, ev):
        """Called on long mouse button press, set selectionMode on if enabled."""
        if self.userChangeSelectionModeEnabled:
            if not self._selectionMode:
                self.setSelectionMode(True)
                return
            elif not self._selection:
                self.setSelectionMode(False)
                return
        super().longMousePressEvent(ev)





# -*- coding: utf-8 -*-
#
# This file is part of the qpageview package.
#
# Copyright (c) 2010 - 2019 by Wilbert Berendsen
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 2
# of the License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
# See http://www.gnu.org/licenses/ for more information.


"""
Highlight rectangular areas inside a View.
"""

import collections
import weakref

from PyQt5.QtCore import QRect, QRectF, QTimer
from PyQt5.QtGui import QPainter, QPen
from PyQt5.QtWidgets import QApplication


class Highlighter:
    """A Highlighter can draw rectangles to highlight e.g. links in a View.

    An instance represents a certain type of highlighting, e.g. of a particular
    style. The paintRects() method is called with a list of rectangles that
    need to be drawn.

    To implement different highlighting behaviour just inherit paintRects().
    The default implementation of paintRects() uses the `color` attribute to get
    the color to use and the `lineWidth` (default: 2) and `radius` (default: 3)
    attributes.

    `lineWidth` specifies the thickness in pixels of the border drawn, `radius`
    specifies the distance in pixels the border is drawn (by default with
    rounded corners) around the area to be highlighted. `color` is set to None
    by default, causing the paintRects method to choose the application's
    palette highlight color.

    """

    lineWidth = 2
    radius = 3
    color = None

    def paintRects(self, painter, rects):
        """Override this method to implement different drawing behaviour."""
        color = self.color if self.color is not None else QApplication.palette().highlight().color()
        pen = QPen(color)
        pen.setWidth(self.lineWidth)
        painter.setPen(pen)
        painter.setRenderHint(QPainter.Antialiasing, True)
        rad = self.radius
        for r in rects:
            r.adjust(-rad, -rad, rad, rad)
            painter.drawRoundedRect(r, rad, rad)


class HighlightViewMixin:
    """Mixin methods vor view.View for highlighting areas.

    This mixin allows for highlighting rectangular areas on pages. You
    can highlight different sets of areas independently, using different
    Highlighter instances.

    Highlighting can be set to stay on forever or to disappear after a
    certain amount of microseconds.

    If desired, the View can be scrolled to show the highlighted areas.
    How the highlighting is drawn is determined by the paintRects() method
    of Highlighter.

    """
    def __init__(self, parent=None, **kwds):
        self._highlights = weakref.WeakKeyDictionary()
        self._defaultHighlighter = None
        super().__init__(parent, **kwds)

    def defaultHighlighter(self):
        """Return a default highlighter, creating it if necessary."""
        if self._defaultHighlighter is None:
            self._defaultHighlighter = Highlighter()
        return self._defaultHighlighter

    def setDefaultHighlighter(self, highlighter):
        """Set a Highlighter to use as the default highlighter."""
        self._defaultHighlighter = highlighter

    def highlightRect(self, areas):
        """Return the bounding rect of the areas."""
        boundingRect = QRect()
        for page, rects in areas.items():
            f = page.mapToPage(1, 1).rect
            pbound = QRect()
            for r in rects:
                pbound |= f(r)
            boundingRect |= pbound.translated(page.pos())
        return boundingRect

    def highlight(self, areas, highlighter=None, msec=0, scroll=False, margins=None, allowKinetic=True):
        """Highlight the areas dict using the given or default highlighter.

        The areas dict maps Page objects to lists of rectangles, where the
        rectangle is a QRectF() inside (0, 0, 1, 1) like the area attribute of
        a Link.

        If the highlighter is not specified, the default highlighter will be
        used.

        If msec > 0, the highlighting will vanish after that many microseconds.

        If scroll is True, the View will be scrolled to show the areas to
        highlight if needed, using View.ensureVisible(highlightRect(areas),
        margins, allowKinetic).

        """
        if highlighter is None:
            highlighter = self.defaultHighlighter()

        if scroll:
            self.ensureVisible(self.highlightRect(areas), margins, allowKinetic)
            if msec:
                msec += self.remainingScrollTime()

        d = weakref.WeakKeyDictionary(areas)
        if msec:
            selfref = weakref.ref(self)
            def clear():
                self = selfref()
                if self:
                    self.clearHighlight(highlighter)
            t = QTimer(singleShot = True, timeout = clear)
            t.start(msec)
        else:
            t = None
        self.clearHighlight(highlighter)
        self._highlights[highlighter] = (d, t)
        self.viewport().update()

    def clearHighlight(self, highlighter=None):
        """Removes the highlighted areas of the given or default highlighter."""
        if highlighter is None:
            highlighter = self.defaultHighlighter()
        try:
            (d, t) = self._highlights[highlighter]
        except KeyError:
            return
        if t is not None: t.stop()
        del self._highlights[highlighter]
        self.viewport().update()

    def isHighlighting(self, highlighter=None):
        """Return True if the given or default highlighter is active."""
        if highlighter is None:
            highlighter = self.defaultHighlighter()
        return highlighter in self._highlights

    def highlightUrls(self, urls, highlighter=None, msec=0, scroll=False, margins=None, allowKinetic=True):
        """Convenience method highlighting the specified urls in the Document.

        The urls argument is a list of urls (str); the other arguments
        are used for calling highlight() on the areas returned by
        getUrlHighlightAreas(urls).

        """
        areas = self.getUrlHighlightAreas(urls)
        if areas:
            self.highlight(areas, highlighter, msec, scroll, margins, allowKinetic)

    def getUrlHighlightAreas(self, urls):
        """Return the areas to highlight all occurrences of the specified URLs.

        The areas are found in the dictionary returned by document().urls().
        URLs that are not in that dictionary are silently skipped.
        If there is no document set this method returns nothing.

        """
        doc = self.document()
        if doc:
            u = doc.urls()
            if u:
                pages = doc.pages()
                areas = collections.defaultdict(list)
                for url in urls:
                    d = u.get(url)
                    if d:
                        for n, linkareas in d.items():
                            rects = []
                            for a in linkareas:
                                r = QRectF()
                                r.setCoords(*a)
                                rects.append(r)
                            areas[pages[n]].extend(rects)
                return areas

    def paintEvent(self, ev):
        """Paint the highlighted areas in the viewport."""
        super().paintEvent(ev)  # first paint the contents
        painter = QPainter(self.viewport())
        for highlighter, (d, t) in self._highlights.items():
            for page, rect in self.pagesToPaint(ev.rect(), painter):
                try:
                    areas = d[page]
                except KeyError:
                    continue
                rectarea = page.mapFromPage(1, 1).rect(rect)
                f = page.mapToPage(1, 1).rect
                rects = [f(area) for area in areas if area & rectarea]
                highlighter.paintRects(painter, rects)





# -*- coding: utf-8 -*-
#
# This file is part of the qpageview package.
#
# Copyright (c) 2019 - 2019 by Wilbert Berendsen
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 2
# of the License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
# See http://www.gnu.org/licenses/ for more information.

"""
Printing facilities for qpageview.
"""


from PyQt5.QtCore import pyqtSignal
from PyQt5.QtGui import QPainter, QTransform
from PyQt5.QtWidgets import QMessageBox, QProgressDialog

from . import backgroundjob


class PrintJob(backgroundjob.Job):
    """Performs a print job in the background.

    Emits the following signals:

        ``progress(pageNumber, num, total)``
            before each Page
        ``finished()``
            when done

    """
    progress = pyqtSignal(int, int, int)

    aborted = False

    def __init__(self, printer, pageList, parent=None):
        """Initialize with a QPrinter object and a list of pages.

        pageList may be a list of two-tuples (num, page). Otherwise, the pages
        are numbered from 1 in the progress message. The pages are copied.

        """
        super().__init__(parent)
        self.printer = printer
        self.setPageList(pageList)

    def setPageList(self, pageList):
        """Set the pagelist to print.

        pageList may be a list of two-tuples (num, page). Otherwise, the pages
        are numbered from 1 in the progress message. The pages are copied.

        """
        self.pageList = []
        for n, page in enumerate(pageList, 1):
            if isinstance(page, tuple):
                pageNum, page = page
            else:
                pageNum = n
            page = page.copy()
            # set zoom to 1.0 so computations based on geometry() are
            # accurate enough
            page.updateSize(page.dpi, page.dpi, 1.0)
            self.pageList.append((pageNum, page))

    def work(self):
        """Paint the pages to the printer in the background."""
        p = self.printer
        p.setFullPage(True)
        painter = QPainter(p)
        for n, (num, page) in enumerate(self.pageList):
            if self.isInterruptionRequested():
                self.aborted = True
                return p.abort()
            self.progress.emit(num, n+1, len(self.pageList))
            if n:
                p.newPage()
            painter.save()
            # center on the page and use scale 100% (TEMP)
            r = p.pageRect()
            m = QTransform()
            m.translate(r.center().x(), r.center().y())
            m.scale(p.logicalDpiX() / page.dpi, p.logicalDpiY() / page.dpi)
            m.rotate(page.rotation * 90)
            m.scale(page.scaleX, page.scaleY)
            m.translate(page.pageWidth / -2, page.pageHeight / -2)
            painter.setTransform(m, True)
            page.print(painter)
            painter.restore()
        return painter.end()


class PrintProgressDialog(QProgressDialog):
    """A simple progress dialog displaying the printing progress."""
    def __init__(self, job, parent=None):
        """Initializes ourselves with the print job and optional parent widget."""
        super().__init__(parent)
        self._job = job
        job.progress.connect(self.showProgress)
        job.finished.connect(self.jobFinished)
        self.canceled.connect(job.requestInterruption)
        self.setMinimumDuration(0)
        self.setRange(0, len(job.pageList))
        self.setLabelText("Preparing to print...")

    def showProgress(self, page, num, total):
        """Called by the job when printing a page."""
        self.setValue(num)
        self.setLabelText("Printing page {page} ({num} of {total})...".format(
                page=page, num=num, total=total))

    def jobFinished(self):
        """Called when the print job has finished."""
        if not self._job.result and not self._job.aborted:
            self.showErrorMessage()
        del self._job
        self.deleteLater()

    def showErrorMessage(self):
        """Reimplement to show a different or translated error message."""
        QMessageBox.warning(self.parent(), "Printing Error",
                    "Could not send the document to the printer.")





# -*- coding: utf-8 -*-
#
# This file is part of the qpageview package.
#
# Copyright (c) 2014 - 2019 by Wilbert Berendsen
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 2
# of the License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
# See http://www.gnu.org/licenses/ for more information.

r"""
A simple module using CUPS to send a document directly to a printer described
by a QPrinter. This is especially useful with PDF documents.

Uses the `cups` module, although it elegantly fails when that module is not
present. The cups module can be found in the pycups package at
https://pypi.org/project/pycups/ .

There are two methods to send a document to a printer:

1. Using the `lp` shell command
2. Using the cups module, which uses libcups to directly contact the server.

This module provides both possibilities.

Use `CmdHandle.create()` to get a CmdHandle, if `lp` is available, or use
`IppHandle.create()` to get a IppHandle, if the cups module is available and a
connection to the server can be established.

A function `handle()` is available; that tries first to get an IppHandle and
then a LprHandle. Usage of this module is this simple::

    import qpageview.cupsprinter

    h = qpageview.cupsprinter.handle()
    if h:
        h.printFile('/path/to/document.pdf')

You can supply a QPrinter instance (that'd be the normal workflow :-)

::

    h = qpageview.cupsprinter.handle(printer)
    if h:
        h.printFile('/path/to/document.pdf')

In this case all options that are set in the QPrinter object will be used
when sending the document to the printer.

If `printFile()` returns True, printing is considered successful. If False,
you can read the `status` and `error` attributes::

    if not h.printFile('/path/to/document.pdf'):
        QMessageBox.warning(None, "Printing failure",
            "There was an error:\n{0} (status: {1})".format(h.error, h.status))

To print a list of files in one job, use `printFiles()`.

"""

import os
import shutil
import subprocess

from PyQt5.QtPrintSupport import QPrintEngine, QPrinter


class Handle:
    """Shared implementation of a handle that can send documents to a printer."""
    def __init__(self, printer=None):
        self._printer = printer

    def setPrinter(self, printer):
        """Use the specified QPrinter."""
        self._printer = printer

    def printer(self):
        """Return the QPrinter given on init, or a new default QPrinter instance."""
        if self._printer == None:
            self._printer = QPrinter()
        return self._printer

    def options(self):
        """Return the dict of CUPS options read from the printer object."""
        return options(self.printer())

    def title(self, filenames):
        """Return a sensible job title based on the list of filenames.

        This method is called when the user did not specify a job title.

        """
        maxlen = 5
        titles = [os.path.basename(f) for f in filenames[:maxlen]]
        more = len(filenames) - maxlen
        if more > 0:
            titles.append("(+{0} more)".format(more))
        return ", ".join(titles)

    def printFile(self, filename, title=None, options=None):
        """Print the file."""
        return self.printFiles([filename], title, options)

    def printFiles(self, filenames, title=None, options=None):
        """Print a list of files.

        If the title is None, the basename of the filename is used. Options may
        be a dictionary of CUPS options.  All keys and values should be
        strings.

        Returns True if the operation was successful. Returns False if there was
        an error; after the call to printFile(), the status and error attributes
        contain the returncode of the operation and the error message.

        """
        if filenames:
            if all(f and not f.isspace() and f != "-" for f in filenames):
                if not title:
                    title = self.title(filenames)
                o = self.options()
                if options:
                    o.update(options)
                printerName = self.printer().printerName()
                self.status, self.error = self._doPrintFiles(printerName, filenames, title, o)
            else:
                self.status, self.error = 2, "Not a valid filename"
        else:
            self.status, self.error = 2, "No filenames specified"
        return self.status == 0

    def _doPrintFiles(self, printerName, filenames, title, options):
        """Implement this to perform the printing.

        Should return a tuple (status, error). If status is 0, the operation is
        considered to be successful. If not, the operation is considered to have
        failed, and the `error` message should contain some more information.

        """
        return 0, ""


class CmdHandle(Handle):
    """Print a document using the `lp` shell command."""
    def __init__(self, command, server="", port=0, user="", printer=None):
        self._command = command
        self._server = server
        self._port = port
        self._user = user
        super().__init__(printer)

    @classmethod
    def create(cls, printer=None, server="", port=0, user="", cmd="lp"):
        """Create a handle to print using a shell command, if available."""
        cmd = shutil.which(cmd)
        if cmd:
            return cls(cmd, server, port, user, printer)

    def _doPrintFiles(self, printerName, filenames, title, options):
        """Print filenames using the `lp` shell command."""
        cmd = [self._command]
        if self._server:
            if self._port:
                cmd.extend(['-h', "{0}:{1}".format(self._server, self._port)])
            else:
                cmd.extend(['-h', self._server])
        if self._user:
            cmd.extend(['-U', self._user])
        cmd.extend(['-d', printerName])
        cmd.extend(['-t', title])
        if options:
            for option, value in options.items():
                cmd.extend(['-o', '{0}={1}'.format(option, value)])
        if any(f.startswith('-') for f in filenames):
            cmd.append('--')
        cmd.extend(filenames)
        try:
            p = subprocess.Popen(cmd, stdin=subprocess.DEVNULL, stderr=subprocess.PIPE)
        except OSError as e:
            return e.errno, e.strerror
        message = p.communicate()[1].decode('UTF-8', 'replace')
        return p.wait(), message


class IppHandle(Handle):
    """Print a document using a connection to the CUPS server."""
    def __init__(self, connection, printer=None):
        super().__init__(printer)
        self._connection = connection

    @classmethod
    def create(cls, printer=None, server="", port=0, user=""):
        """Return a handle to print using a connection to the (local) CUPS server, if available."""
        try:
            import cups
        except ImportError:
            return
        cups.setServer(server or "")
        cups.setPort(port or 0)
        cups.setUser(user or "")
        try:
            c = cups.Connection()
        except RuntimeError:
            return
        h = cls(c, printer)
        if h.printer().printerName() in c.getPrinters():
            return h

    def _doPrintFiles(self, printerName, filenames, title, options):
        """Print filenames using a connection to the CUPS server."""
        import cups
        # cups.Connection.printFiles() behaves flaky: version 1.9.74 can
        # silently fail (without returning an error), and after having fixed
        # that, there are strange error messages on some options.
        # Therefore we use cups.printFile() for every file.
        for filename in filenames:
            try:
                self._connection.printFile(printerName, filename, title, options)
            except cups.IPPError as err:
                return err.args
        return 0, ""


def handle(printer=None, server="", port=0, user=""):
    """Return the first available handle to print a document to a CUPS server."""
    return (IppHandle.create(printer, server, port, user) or
            CmdHandle.create(printer, server, port, user))


def options(printer):
    """Return the dict of CUPS options read from the QPrinter object."""
    o = {}

    # cups options that can be set in QPrintDialog on unix
    # I found this in qt5/qtbase/src/printsupport/kernel/qcups.cpp.
    # Esp. options like page-set even/odd do make sense.
    props = printer.printEngine().property(0xfe00)
    if props and isinstance(props, list) and len(props) % 2 == 0:
        for key, value in zip(props[0::2], props[1::2]):
            if value and isinstance(key, str) and isinstance(value, str):
                o[key] = value

    o['copies'] = format(printer.copyCount())
    if printer.collateCopies():
        o['collate'] = 'true'

    # TODO: in Qt5 >= 5.11 page-ranges support is more fine-grained!
    if printer.printRange() == QPrinter.PageRange:
        o['page-ranges'] = '{0}-{1}'.format(printer.fromPage(), printer.toPage())

    # page order
    if printer.pageOrder() == QPrinter.LastPageFirst:
        o['outputorder'] = 'reverse'

    # media size
    media = []
    size = printer.paperSize()
    if size == QPrinter.Custom:
        media.append('Custom.{0}x{1}mm'.format(printer.heightMM(), printer.widthMM()))
    elif size in PAGE_SIZES:
        media.append(PAGE_SIZES[size])

    # media source
    source = printer.paperSource()
    if source in PAPER_SOURCES:
        media.append(PAPER_SOURCES[source])

    if media:
        o['media'] = ','.join(media)

    # page margins
    if printer.printEngine().property(QPrintEngine.PPK_PageMargins):
        left, top, right, bottom = printer.getPageMargins(QPrinter.Point)
        o['page-left'] = format(left)
        o['page-top'] = format(top)
        o['page-right'] = format(right)
        o['page-bottom'] = format(bottom)

    # orientation
    landscape = printer.orientation() == QPrinter.Landscape
    if landscape:
        o['landscape'] = 'true'

    # double sided
    duplex = printer.duplex()
    o['sides'] = (
        'two-sided-long-edge' if duplex == QPrinter.DuplexLongSide or
                        (duplex == QPrinter.DuplexAuto and not landscape) else
        'two-sided-short-edge' if duplex == QPrinter.DuplexShortSide or
                        (duplex == QPrinter.DuplexAuto and landscape) else
        'one-sided')

    # grayscale
    if printer.colorMode() == QPrinter.GrayScale:
        o['print-color-mode'] = 'monochrome'

    return o


def clearPageSetSetting(printer):
    """Remove 'page-set' even/odd cups options from the printer's CUPS options.

    Qt's QPrintDialog fails to reset the 'page-set' option back to 'all pages',
    so a previous value (even or odd) could remain in the print options, even
    if the user has selected All Pages in the print dialog.

    This function clears the page-set setting from the cups options. If the
    user selects or has selected even or odd pages, it will be added again by
    the dialog.

    So call this function on a QPrinter, just before showing a QPrintDialog.

    """
    # see qt5/qtbase/src/printsupport/kernel/qcups.cpp
    opts = printer.printEngine().property(0xfe00)
    if opts and isinstance(opts, list) and len(opts) % 2 == 0:
        try:
            i = opts.index('page-set')
        except ValueError:
            return
        if i % 2 == 0:
            del opts[i:i+2]
            printer.printEngine().setProperty(0xfe00, opts)


PAGE_SIZES = {
    QPrinter.A0: "A0",
    QPrinter.A1: "A1",
    QPrinter.A2: "A2",
    QPrinter.A3: "A3",
    QPrinter.A4: "A4",
    QPrinter.A5: "A5",
    QPrinter.A6: "A6",
    QPrinter.A7: "A7",
    QPrinter.A8: "A8",
    QPrinter.A9: "A9",
    QPrinter.B0: "B0",
    QPrinter.B1: "B1",
    QPrinter.B10: "B10",
    QPrinter.B2: "B2",
    QPrinter.B3: "B3",
    QPrinter.B4: "B4",
    QPrinter.B5: "B5",
    QPrinter.B6: "B6",
    QPrinter.B7: "B7",
    QPrinter.B8: "B8",
    QPrinter.B9: "B9",
    QPrinter.C5E: "C5",         # Correct Translation?
    QPrinter.Comm10E: "Comm10", # Correct Translation?
    QPrinter.DLE: "DL",         # Correct Translation?
    QPrinter.Executive: "Executive",
    QPrinter.Folio: "Folio",
    QPrinter.Ledger: "Ledger",
    QPrinter.Legal: "Legal",
    QPrinter.Letter: "Letter",
    QPrinter.Tabloid: "Tabloid",
}

PAPER_SOURCES = {
    QPrinter.Cassette: "Cassette",
    QPrinter.Envelope: "Envelope",
    QPrinter.EnvelopeManual: "EnvelopeManual",
    QPrinter.FormSource: "FormSource",
    QPrinter.LargeCapacity: "LargeCapacity",
    QPrinter.LargeFormat: "LargeFormat",
    QPrinter.Lower: "Lower",
    QPrinter.MaxPageSource: "MaxPageSource",
    QPrinter.Middle: "Middle",
    QPrinter.Manual: "Manual",
    QPrinter.OnlyOne: "OnlyOne",
    QPrinter.Tractor: "Tractor",
    QPrinter.SmallFormat: "SmallFormat",
}




# -*- coding: utf-8 -*-
#
# This file is part of the qpageview package.
#
# Copyright (c) 2019 - 2019 by Wilbert Berendsen
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 2
# of the License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
# See http://www.gnu.org/licenses/ for more information.

"""
SidebarView, a special View with miniatures to use as a sidebar for a View.

Automatically displays all pages in a view in small size, and makes it easier
to browse large documents.

"""

from PyQt5.QtCore import QEvent, QMargins, QRect, Qt
from PyQt5.QtGui import QPainter

from . import constants
from . import layout
from . import selector
from . import view
from . import util



class SidebarView(selector.SelectorViewMixin, util.LongMousePressMixin, view.View):
    """A special View with miniatures to use as a sidebar for a View.

    Automatically displays all pages in a view in small size, and makes it
    easier to browse large documents. Use setView() to connect a View, and
    it automatically shows the pages, also when the view is changed.

    """

    MAX_ZOOM = 1.0
    pagingOnScrollEnabled = False
    wheelZoomingEnabled = False
    firstPageNumber = 1
    scrollupdatespersec = 100

    autoOrientationEnabled = True

    def __init__(self, parent=None, **kwds):
        super().__init__(parent, **kwds)
        self._view = None
        self.setOrientation(constants.Vertical)
        self.pageLayout().spacing = 1
        self.pageLayout().setMargins(QMargins(0, 0, 0, 0))
        self.pageLayout().setPageMargins(QMargins(4, 4, 4, 20))
        self.setLayoutFontHeight()
        self.currentPageNumberChanged.connect(self.viewport().update)

    def setOrientation(self, orientation):
        """Reimplemented to also set the corresponding view mode."""
        super().setOrientation(orientation)
        if orientation == constants.Vertical:
            self.setViewMode(constants.FitWidth)
        else:
            self.setViewMode(constants.FitHeight)

    def setLayoutFontHeight(self):
        """Reads the current font height and reserves enough space in the layout."""
        self.pageLayout().pageMargins().setBottom(self.fontMetrics().height())
        self.updatePageLayout()

    def setView(self, view):
        """Connects to a View, or disconnects the current view if view is None."""
        if view is not self._view:
            if self._view:
                self.currentPageNumberChanged.disconnect(self._view.setCurrentPageNumber)
                self._view.currentPageNumberChanged.disconnect(self.slotCurrentPageNumberChanged)
                self._view.pageLayoutUpdated.disconnect(self.slotLayoutUpdated)
                self.clear()
            self._view = view
            if view:
                self.slotLayoutUpdated()
                self.setCurrentPageNumber(view.currentPageNumber())
                self.currentPageNumberChanged.connect(view.setCurrentPageNumber)
                view.currentPageNumberChanged.connect(self.slotCurrentPageNumberChanged)
                view.pageLayoutUpdated.connect(self.slotLayoutUpdated)

    def slotLayoutUpdated(self):
        """Called when the layout of the connected view is updated."""
        self.pageLayout()[:] = (p.copy(self) for p in self._view.pageLayout())
        self.pageLayout().rotation = self._view.pageLayout().rotation
        self.updatePageLayout()

    def slotCurrentPageNumberChanged(self, num):
        """Called when the page number in the connected view changes.

        Does not scroll but updates the current page mark in our View.

        """
        self._currentPageNumber = num
        self.viewport().update()

    def paintEvent(self, ev):
        """Reimplemented to print page numbers and a selection box."""
        painter = QPainter(self.viewport())
        layout = self.pageLayout()
        for p, rect in self.pagesToPaint(ev.rect(), painter):
            ## draw selection background on current page
            if p is self.currentPage():
                bg = rect + layout.pageMargins()
                painter.fillRect(bg, self.palette().highlight())
                painter.setPen(self.palette().highlightedText().color())
            else:
                painter.setPen(self.palette().text().color())
            # draw text
            textr = QRect(rect.x(), rect.bottom(), rect.width(), layout.pageMargins().bottom())
            painter.drawText(textr, Qt.AlignCenter, str(layout.index(p) + self.firstPageNumber))
        super().paintEvent(ev)

    def wheelEvent(self, ev):
        """Reimplemented to page instead of scroll."""
        if ev.angleDelta().y() > 0:
            self.gotoPreviousPage()
        elif ev.angleDelta().y() < 0:
            self.gotoNextPage()

    def keyPressEvent(self, ev):
        """Reimplemented to page instead of scroll."""
        if ev.key() in (Qt.Key_PageDown, Qt.Key_Down):
            self.gotoNextPage()
        elif ev.key() in (Qt.Key_PageUp, Qt.Key_Up):
            self.gotoPreviousPage()
        elif ev.key() == Qt.Key_End:
            self.setCurrentPageNumber(self.pageCount())
        elif ev.key() == Qt.Key_Home:
            self.setCurrentPageNumber(1)
        else:
            super().keyPressEvent(ev)

    def resizeEvent(self, ev):
        """Reimplemented to auto-change the orientation if desired."""
        super().resizeEvent(ev)
        if self.autoOrientationEnabled:
            s = ev.size()
            if s.width() > s.height() and self.orientation() == constants.Vertical:
                self.setOrientation(constants.Horizontal)
            elif s.width() < s.height() and self.orientation() == constants.Horizontal:
                self.setOrientation(constants.Vertical)

    def changeEvent(self, ev):
        """Reimplemented to set the correct font height for the page numbers."""
        super().changeEvent(ev)
        if ev.type() in (QEvent.ApplicationFontChange, QEvent.FontChange):
            self.setLayoutFontHeight()





# -*- coding: utf-8 -*-
#
# This file is part of the qpageview package.
#
# Copyright (c) 2019 - 2019 by Wilbert Berendsen
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 2
# of the License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
# See http://www.gnu.org/licenses/ for more information.

"""
A View mixin class that draws a nice drop shadow around all pages.
"""

from PyQt5.QtCore import QPoint, Qt
from PyQt5.QtGui import QColor, QPainter, QPen


class ShadowViewMixin:
    """Mixin class that draws a drop shadow around every Page.

    Drawing the drop shadow can be turned off by setting dropShadowEnabled
    to False.

    """

    dropShadowEnabled = True

    def paintEvent(self, ev):
        if self.dropShadowEnabled:
            width = round(self._pageLayout.spacing / 2.0)
            # make the rect slightly larger, so we "see" shadow of pages that
            # would be outside view normally.
            rect = ev.rect().adjusted(-width, -width, width // 2, width // 2)
            painter = QPainter(self.viewport())
            for page, rect in self.pagesToPaint(rect, painter):
                self.drawDropShadow(page, painter, width)
        super().paintEvent(ev)      # then draw the contents

    def drawDropShadow(self, page, painter, width):
        """Draw a drop shadow of width pixels around the Page.

        The painter is already translated to the topleft corner of the Page.

        """
        width = round(width)
        rect = page.rect().adjusted(width // 2, width // 2, 0, 0)
        color = QColor(Qt.black)
        pen = QPen()
        pen.setWidth(1)
        pen.setJoinStyle(Qt.MiterJoin)
        for i in range(width):
            f = (width-i)/width
            color.setAlpha(int(200**f + 55*f))
            pen.setColor(color)
            painter.setPen(pen)
            painter.drawRect(rect.adjusted(-i, -i, i, i))




